import numpy as np


class GradientOptimizer(object):
    def __init__(self, learning_rate=0.1):
        self.learning_rate = learning_rate
        self.dW = None
        self.db = None

    def delta(self, grad_weights, grad_biases):
        self.grad_weights = grad_weights
        self.grad_biases = grad_biases
        return self.learning_rate * self.grad_weights, self.learning_rate * self.grad_biases


class MomentumOptimizer(GradientOptimizer):
    def __init__(self, learning_rate=0.1, beta=0.9):
        GradientOptimizer.__init__(self, learning_rate)
        assert (beta < 1)
        self.beta = beta

        self.initialized = False

        self.vW = None
        self.vb = None

    def delta(self, grad_weights, grad_biases):
        self.grad_weights = grad_weights
        self.grad_biases = grad_biases

        if not self.initialized:
            self.vW = np.zeros(grad_weights.shape)
            self.vb = np.zeros(grad_biases.shape)
            self.initialized = True

        self.vW = (self.beta * self.vW + (1 - self.beta) * self.grad_weights)
        self.vb = (self.beta * self.vb + (1 - self.beta) * self.grad_biases)
        return self.learning_rate * self.vW, self.learning_rate * self.vb


class RmspropOptimizer(object):
    def __init__(self, learning_rate=0.01, beta=0.9, epsilon=1e-8):
        GradientOptimizer.__init__(self, learning_rate)
        assert (beta < 1)
        self.beta = beta
        self.epsilon = epsilon

        self.initialized = False

        self.sW = None
        self.sb = None

    def delta(self, grad_weights, grad_biases):
        self.grad_weights = grad_weights
        self.grad_biases = grad_biases

        if not self.initialized:
            self.sW = np.zeros(grad_weights.shape)
            self.sb = np.zeros(grad_biases.shape)
            self.initialized = True

        self.sW = self.beta * self.sW + (1 - self.beta) * np.power(self.grad_weights, 2)
        self.sb = self.beta * self.sb + (1 - self.beta) * np.power(self.grad_biases, 2)

        deltaW = self.learning_rate * self.grad_weights / (np.sqrt(self.sW) + self.epsilon)
        deltaB = self.learning_rate * self.grad_biases / (np.sqrt(self.sb) + self.epsilon)
        return deltaW, deltaB


class AdamOptimizer(object):
    def __init__(self, learning_rate=0.001, beta1=0.9, beta2=0.999, epsilon=1e-8):
        GradientOptimizer.__init__(self, learning_rate)
        assert (beta1 < 1)
        assert (beta2 < 1)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon

        self.t = 0
        self.initialized = False

        self.vW = None
        self.sW = None
        self.vb = None
        self.sb = None

        self.vW_corrected = None
        self.sW_corrected = None
        self.vb_corrected = None
        self.sb_corrected = None

    def delta(self, grad_weights, grad_biases):
        self.grad_weights = grad_weights
        self.grad_biases = grad_biases

        if not self.initialized:
            self.vW = np.zeros(grad_weights.shape)
            self.sW = np.zeros(grad_weights.shape)
            self.vb = np.zeros(grad_biases.shape)
            self.sb = np.zeros(grad_biases.shape)
            self.initialized = True

        self.t = self.t + 1

        self.vW = self.beta1 * self.vW + (1 - self.beta1) * self.grad_weights
        self.sW = self.beta2 * self.sW + (1 - self.beta2) * np.power(self.grad_weights, 2)
        self.vb = self.beta1 * self.vb + (1 - self.beta1) * self.grad_biases
        self.sb = self.beta2 * self.sb + (1 - self.beta2) * np.power(self.grad_biases, 2)

        self.vW_corrected = self.vW / (1 - self.beta1 ** self.t)
        self.sW_corrected = self.sW / (1 - self.beta2 ** self.t)
        self.vb_corrected = self.vb / (1 - self.beta1 ** self.t)
        self.sb_corrected = self.sb / (1 - self.beta2 ** self.t)

        deltaW = self.learning_rate * self.vW_corrected / (np.sqrt(self.sW_corrected) + self.epsilon)
        deltaB = self.learning_rate * self.vb_corrected / (np.sqrt(self.sb_corrected) + self.epsilon)

        return deltaW, deltaB



import matplotlib.pyplot as plt


class Model(object):
    def __init__(self):
        self.costs = []

    def activate(self, X):
        return X

    def get_cost(self, X, Y):
        return 0

    def plot_costs(self, title=None, xlabel="epochs", ylabel="cost", costs=None):
        if costs is None:
            costs = self.costs
        plt.plot(costs)
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.title(title)
        plt.show()





import matplotlib.pyplot as plt
import numpy as np
import sklearn
import sklearn.datasets
import sklearn.linear_model
import sklearn.model_selection
import h5py
import scipy.io


# def plot_decision_boundary(model, X, y):
#     # Set min and max values and give it some padding
#     x_min, x_max = X[0, :].min() - 1, X[0, :].max() + 1
#     y_min, y_max = X[1, :].min() - 1, X[1, :].max() + 1
#     h = 0.01
#     # Generate a grid of points with distance h between them
#     xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))
#     # Predict the function value for the whole grid
#     Z = model(np.c_[xx.ravel(), yy.ravel()])
#     Z = Z.reshape(xx.shape)
#     # Plot the contour and training examples
#     plt.contourf(xx, yy, Z, cmap=plt.cm.Spectral)
#     plt.ylabel('x2')
#     plt.xlabel('x1')
#     plt.scatter(X[0, :], X[1, :], c=y[0, :], cmap=plt.cm.Spectral)


# def plot_decision_boundary(model, X, y):
#     # Set min and max values and give it some padding
#     x_min, x_max = X[0, :].min() - 1, X[0, :].max() + 1
#     y_min, y_max = X[1, :].min() - 1, X[1, :].max() + 1
#     h = 0.01
#     # Generate a grid of points with distance h between them
#     xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))
#     # Predict the function value for the whole grid
#     Z = model(np.c_[xx.ravel(), yy.ravel()])
#     Z = Z.reshape(xx.shape)
#     # Plot the contour and training examples
#     plt.contourf(xx, yy, Z, cmap=plt.cm.Spectral)
#     plt.ylabel('x2')
#     plt.xlabel('x1')
#     plt.scatter(X[0, :], X[1, :], c=y[0, :], cmap=plt.cm.Spectral)
#     plt.show()

# Helper function to plot a decision boundary.
# If you don't fully understand this function don't worry, it just generates the contour plot below.
# def plot_decision_boundary(pred_func, X, y):
#     # Set min and max values and give it some padding
#     x_min, x_max = X[:, 0].min() - .5, X[:, 0].max() + .5
#     y_min, y_max = X[:, 1].min() - .5, X[:, 1].max() + .5
#     h = 0.01
#     # Generate a grid of points with distance h between them
#     xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))
#     # Predict the function value for the whole gid
#     Z = pred_func(np.c_[xx.ravel(), yy.ravel()])
#     Z = Z.reshape(xx.shape)
#     # Plot the contour and training examples
#     plt.contourf(xx, yy, Z, cmap=plt.cm.Spectral)
#     plt.scatter(X[:, 0], X[:, 1], c=y[0, :], cmap=plt.cm.Spectral)

def map_feature(X, degree=2):
    '''
    generate the composition of features of X.
    :param X:
    :param degree:
    :return:
    '''
    assert(X.shape[0]<=2)
    if X.shape[0]==2:
        x1 = X[0,:]
        x2 = X[1,:]
        m = np.sum(np.array([i+1 for i in range(1, degree+1)]))
        out = np.zeros((m, X.shape[1]))
        k = 0
        for i in range(1, degree+1):
            for j in range(0, i+1):
                out[k, :] = np.multiply(np.power(x1, i-j), np.power(x2, j))
                k = k + 1
        return out
    else:
        x1 = X[0, :]
        out = np.zeros((degree, X.shape[1]))
        for i in range(0, degree):
            out[i, :] = np.power(x1, i+1)
        return out


def plot_decision_boundary(model, X, y):
    # Set min and max values and give it some padding
    x_min, x_max = X[0, :].min() - 1, X[0, :].max() + 1
    y_min, y_max = X[1, :].min() - 1, X[1, :].max() + 1
    h = 0.01
    # Generate a grid of points with distance h between them
    xx, yy = np.meshgrid(np.arange(x_min, x_max, h), np.arange(y_min, y_max, h))
    # Predict the function value for the whole grid
    Z = model(np.c_[xx.ravel(), yy.ravel()])
    Z = Z.reshape(xx.shape)
    # Plot the contour and training examples
    plt.contourf(xx, yy, Z, cmap=plt.cm.Spectral)
    plt.ylabel('x2')
    plt.xlabel('x1')
    plt.scatter(X[0, :], X[1, :], c=y[0, :], cmap=plt.cm.Spectral)
    plt.show()


# map_feature
def plot_decision_boundary_map_feature(predictor, X, y, title="decision boundary", degree=2 ):
    plt.title(title)
    axes = plt.gca()
    min = X.min(axis=1)
    max = X.max(axis=1)
    range = max - min
    min = min - 0.1 * range
    max = max + 0.1 * range
    axes.set_xlim((min[0], max[0]))
    axes.set_ylim((min[1], max[1]))
    plot_decision_boundary(lambda x: predictor.predict(map_feature(x.T, degree)), X, y)


def load_sin_line_dataset(data_path="../../data/sin_line.txt", delimiter="\t", random_state=3, test_size=0.25):
    '''
     ex1data2.txt contains a training set of housing prices in Portland, Oregon.
     The first column is the size of the house (in square feet), the second column
     is the number of bedrooms, and the third column is the price of the house.
    :param data_path:
    :param delimiter:
    :return:
    '''
    data = np.loadtxt(data_path, delimiter=delimiter)
    X = data[:, :-1]
    y = data[:, -1:]
    train_X, test_X, train_y, test_y = sklearn.model_selection.train_test_split(
        X, y, test_size=test_size, random_state=random_state)
    return train_X.T, train_y.T, test_X.T, test_y.T


def load_flat_dataset(data_path="../../data/ex1data1.txt", delimiter=","):
    '''
     ex1data2.txt contains a training set of housing prices in Portland, Oregon.
     The first column is the size of the house (in square feet), the second column
     is the number of bedrooms, and the third column is the price of the house.
    :param data_path:
    :param delimiter:
    :return:
    '''
    data = np.loadtxt(data_path, delimiter=delimiter)
    X = data[:, :-1]
    y = data[:, -1:]
    return X.T, y.T


def load_regression_dataset(n_samples=400, n_features=1, bias=1, noise=5, random_state=3, test_size=0.25):
    X, y = sklearn.datasets.make_regression(n_samples=n_samples, n_features=n_features, n_targets=1, bias=bias,
                                            coef=False, noise=noise, random_state=random_state)
    train_X, test_X, train_y, test_y = sklearn.model_selection.train_test_split(
        X, y, test_size = test_size, random_state = random_state)
    return train_X, train_y, test_X, test_y


def load_regression_dataset_coef(n_samples=400, n_features=1, bias=1, noise=5, random_state=3, test_size=0.25, b=5):
    X, y, w = sklearn.datasets.make_regression(n_samples=n_samples, n_features=n_features, n_targets=1, bias=bias,
                                            coef=True, noise=noise, random_state=random_state)
    train_X, test_X, train_y, test_y = sklearn.model_selection.train_test_split(
        X, y, test_size = test_size, random_state = random_state)
    return train_X, train_y+b, test_X, test_y+b, w, b


def load_happy_dataset(train_data_path='../../data/train_happy.h5', test_data_path='../../data/test_happy.h5'):
    train_dataset = h5py.File(train_data_path, "r")
    train_set_x_orig = np.array(train_dataset["train_set_x"][:])  # your train set features
    train_set_y_orig = np.array(train_dataset["train_set_y"][:])  # your train set labels

    test_dataset = h5py.File(test_data_path, "r")
    test_set_x_orig = np.array(test_dataset["test_set_x"][:])  # your test set features
    test_set_y_orig = np.array(test_dataset["test_set_y"][:])  # your test set labels

    classes = np.array(test_dataset["list_classes"][:])  # the list of classes

    train_set_y_orig = train_set_y_orig.reshape((1, train_set_y_orig.shape[0]))
    test_set_y_orig = test_set_y_orig.reshape((1, test_set_y_orig.shape[0]))

    return train_set_x_orig, train_set_y_orig, test_set_x_orig, test_set_y_orig, classes


def load_boston_dataset(test_size=0.25, random_state=3):
    '''
    -Origin
        The origin of the boston housing data is Natural.
    -Usage
        This dataset may be used for Assessment.
    -Number of Cases
        The dataset contains a total of 506 cases.
    -Order
        The order of the cases is mysterious.
    -Variables
        There are 14 attributes in each case of the dataset. They are:
            CRIM - per capita crime rate by town
            ZN - proportion of residential land zoned for lots over 25,000 sq.ft.
            INDUS - proportion of non-retail business acres per town.
            CHAS - Charles River dummy variable (1 if tract bounds river; 0 otherwise)
            NOX - nitric oxides concentration (parts per 10 million)
            RM - average number of rooms per dwelling
            AGE - proportion of owner-occupied units built prior to 1940
            DIS - weighted distances to five Boston employment centres
            RAD - index of accessibility to radial highways
            TAX - full-value property-tax rate per $10,000
            PTRATIO - pupil-teacher ratio by town
            B - 1000(Bk - 0.63)^2 where Bk is the proportion of blacks by town
            LSTAT - % lower status of the population
            MEDV - Median value of owner-occupied homes in $1000's
    :return:
    '''
    boston = sklearn.datasets.load_boston()
    train_X, test_X, train_y, test_y = sklearn.model_selection.train_test_split(
        boston.data, boston.target, test_size = test_size, random_state = random_state)
    return train_X, train_y, test_X, test_y


def load_planar_dataset():
    np.random.seed(1)
    m = 400  # number of examples
    N = int(m / 2)  # number of points per class
    D = 2  # dimensionality
    X = np.zeros((m, D))  # data matrix where each row is a single example
    Y = np.zeros((m, 1), dtype='uint8')  # labels vector (0 for red, 1 for blue)
    a = 4  # maximum ray of the flower

    for j in range(2):
        ix = range(N * j, N * (j + 1))
        t = np.linspace(j * 3.12, (j + 1) * 3.12, N) + np.random.randn(N) * 0.2  # theta
        r = a * np.sin(4 * t) + np.random.randn(N) * 0.2  # radius
        X[ix] = np.c_[r * np.sin(t), r * np.cos(t)]
        Y[ix] = j

    X = X.T
    Y = Y.T

    return X, Y


def load_petal_dataset(num_example=400, random_state=3):
    np.random.seed(1)
    m = num_example  # number of examples
    N = int(m / 2)  # number of points per class
    D = 2  # dimensionality
    X = np.zeros((m, D))  # data matrix where each row is a single example
    Y = np.zeros((m, 1), dtype='uint8')  # labels vector (0 for red, 1 for blue)
    a = 4  # maximum ray of the flower

    for j in range(2):
        ix = range(N * j, N * (j + 1))
        t = np.linspace(j * 3.12, (j + 1) * 3.12, N) + np.random.randn(N) * 0.2  # theta
        r =  np.sin(4 * t) + np.random.randn(N) * 0.2 /a # radius
        X[ix] = np.c_[r * np.sin(t), r * np.cos(t)]
        Y[ix] = j


    train_X, test_X, train_y, test_y = sklearn.model_selection.train_test_split(
        X, Y, test_size=0.25, random_state=random_state)

    return train_X.T, train_y.T, test_X.T, test_y.T


def load_extra_datasets():
    N = 200
    noisy_circles = sklearn.datasets.make_circles(n_samples=N, factor=.5, noise=.3)
    noisy_moons = sklearn.datasets.make_moons(n_samples=N, noise=.2)
    blobs = sklearn.datasets.make_blobs(n_samples=N, random_state=5, n_features=2, centers=6)
    gaussian_quantiles = sklearn.datasets.make_gaussian_quantiles(mean=None, cov=0.5, n_samples=N, n_features=2,
                                                                  n_classes=2, shuffle=True, random_state=None)
    no_structure = np.random.rand(N, 2), np.random.rand(N, 2)

    return noisy_circles, noisy_moons, blobs, gaussian_quantiles, no_structure


def load_dataset():
    np.random.seed(1)
    train_X, train_Y = sklearn.datasets.make_circles(n_samples=300, noise=.05)
    np.random.seed(2)
    test_X, test_Y = sklearn.datasets.make_circles(n_samples=100, noise=.05)
    # Visualize the data
    plt.scatter(train_X[:, 0], train_X[:, 1], c=train_Y, s=40, cmap=plt.cm.Spectral);
    train_X = train_X.T
    train_Y = train_Y.reshape((1, train_Y.shape[0]))
    test_X = test_X.T
    test_Y = test_Y.reshape((1, test_Y.shape[0]))
    return train_X, train_Y, test_X, test_Y


def load_circle_dataset(n_train=300, seed_train=1, noise_train=0.05, n_test=100, seed_test=2, noise_test=0.05):
    np.random.seed(seed_train)
    train_X, train_Y = sklearn.datasets.make_circles(n_samples=n_train, noise=noise_train)
    np.random.seed(seed_test)
    test_X, test_Y = sklearn.datasets.make_circles(n_samples=n_test, noise=noise_test)
    train_X = train_X.T
    train_Y = train_Y.reshape((1, train_Y.shape[0]))
    test_X = test_X.T
    test_Y = test_Y.reshape((1, test_Y.shape[0]))
    return train_X, train_Y, test_X, test_Y


def load_image_data(train_data_path='../../data/train_catvnoncat.h5', test_data_path='../../data/test_catvnoncat.h5'):

    print('load ' + train_data_path)
    train_dataset = h5py.File(train_data_path, "r")
    train_set_x_orig = np.array(train_dataset["train_set_x"][:])  # your train set features
    train_set_y_orig = np.array(train_dataset["train_set_y"][:])  # your train set labels

    print('load ' + test_data_path)
    test_dataset = h5py.File(test_data_path, "r")
    test_set_x_orig = np.array(test_dataset["test_set_x"][:])  # your test set features
    test_set_y_orig = np.array(test_dataset["test_set_y"][:])  # your test set labels

    classes = np.array(test_dataset["list_classes"][:])  # the list of classes

    train_set_y_orig = train_set_y_orig.reshape((train_set_y_orig.shape[0], 1))
    test_set_y_orig = test_set_y_orig.reshape((test_set_y_orig.shape[0], 1))

    return train_set_x_orig, train_set_y_orig, test_set_x_orig, test_set_y_orig, classes


def initialize_image_data(X, y=None):
    m = X.shape[0]
    X = X.reshape((m, -1)).T
    if y is not None: y = y.T
    return X/255.0, y


def load_moon_dataset():
    np.random.seed(3)
    train_X, train_Y = sklearn.datasets.make_moons(n_samples=300, noise=.2)  # 300 #0.2
    # Visualize the data

    train_X = train_X.T
    train_Y = train_Y.reshape((1, train_Y.shape[0]))

    return train_X, train_Y


def load_football_dataset(file_path='../../data/football.mat'):
    data = scipy.io.loadmat(file_path)
    train_X = data['X'].T
    train_Y = data['y'].T
    test_X = data['Xval'].T
    test_Y = data['yval'].T

    return train_X, train_Y, test_X, test_Y


if __name__ == "__main__":
    # load_image_data()
    # load_football_dataset()
    # load_image_data()
    # train_X, train_y, test_X, test_y = load_boston_dataset()
    # print(train_X.shape)
    # print(train_y.shape)
    # print(test_X.shape)
    # print(test_y.shape)
    train_X, train_y, test_X, test_y = load_regression_dataset()
    print(train_X.shape)
    print(train_y.shape)
    print(test_X.shape)
    print(test_y.shape)


import numpy as np
import math
import matplotlib.pyplot as plt


def rotate(theta, alpha):
    A = np.array([[math.cos(theta), -math.sin(theta)],
                  [math.sin(theta), math.cos(theta)]])
    beta = A.dot(alpha)
    return beta, A


theta = 30/180*math.pi
alpha = np.array([math.sqrt(2)/2, math.sqrt(2)/2])
beta, A  = rotate(theta, alpha)


plt.axis('scaled')
plt.xlim([-1, 1])
plt.ylim([0, 1])
plt.xticks(np.arange(-1, 1, 0.5))
plt.yticks(np.arange(0, 1, 0.5))
plt.grid(linestyle='-.')

plt.quiver([0], [0], A[0,:], A[1,:], angles='xy', scale_units='xy',  scale=1)
plt.text(*A[:,0], '$a_1$', size=12)
plt.text(*A[:,1], '$a_2$', size=12)

plt.quiver([0], [0], alpha[0], alpha[1], angles='xy', scale_units='xy',  scale=1, color='b')
plt.quiver([0], [0], beta[0], beta[1], angles='xy', scale_units='xy',  scale=1, color='g')
plt.text(*alpha, '$\\alpha$', size=12)
plt.text(*beta, '$\\beta$', size=12)

plt.show()


import unittest
from qbz95.ml.linear_model import *
from qbz95.ml.utils import *
from qbz95.ml.planar_utils import *
from qbz95.ml.predictor import Predictor

# logger.setLevel(level=logging.DEBUG)


class TestLinearRegression(unittest.TestCase):
    def test_compute_propagate(self):

        train_X, train_y = load_flat_dataset(data_path="../../../data/ex1data1.txt",)
        model = LinearRegression(num_features=train_X.shape[0],
                                 optimizer=GradientOptimizer(0.01),
                                 iterator=NumberIterator(50))
        model.propagate(train_X, train_y)

        print("train_X.shape={}".format(train_X.shape))
        print("train_y.shape={}".format(train_y.shape))
        print("cost={}".format(model.cost))

        np.testing.assert_almost_equal(model.cost, 32.07273388)

    def test_fit(self):

        train_X, train_y = load_flat_dataset(data_path="../../../data/ex1data1.txt",)
        model = LinearRegression(num_features=train_X.shape[0],
                                 optimizer=GradientOptimizer(0.01),
                                 iterator=NumberIterator(1500))
        model.fit(train_X, train_y, print_cost=True, print_num_epoch=100)
        print("cost={}".format(model.cost))
        print("W={}".format(model.W))
        print("b={}".format(model.b))

        np.testing.assert_array_almost_equal(model.W, [[1.166362350]])
        np.testing.assert_almost_equal(model.b, -3.630291439)
        np.testing.assert_almost_equal(model.cost, 4.483388257)

        predict1 = Predictor(model).predict(np.array([3.5]))
        print("predict1={}".format(predict1))
        np.testing.assert_almost_equal(predict1, 0.4519767868)

        predict2 = Predictor(model).predict(np.array([7]))
        print("predict2={}".format(predict2))
        np.testing.assert_almost_equal(predict2, 4.5342450129)

    def test_normal_equation(self):
        train_X, train_y = load_flat_dataset("../../../data/ex1data2.txt")

        W, b = LinearRegression.normal_equation(train_X, train_y)

        print("W={}".format(W))
        print("b={}".format(b))

        np.testing.assert_array_almost_equal(W, [[139.210674, -8738.019112]])
        np.testing.assert_almost_equal(b, 89597.909542798)

        train_X, mu, sigma = feature_normalize(train_X)
        print(train_X.shape)
        print("mu={}".format(mu))
        print("sigma={}".format(sigma))
        W, b = LinearRegression.normal_equation(train_X, train_y)
        print("W={}".format(W))
        print("b={}".format(b))

        #revert actual W, b
        b = b - np.dot(W, mu/sigma)
        W = W/sigma.T
        print("W={}".format(W))
        print("b={}".format(b))

        np.testing.assert_array_almost_equal(W, [[139.210674, -8738.019112]])
        np.testing.assert_almost_equal(b, 89597.909542798)


    def test_interate(self):
        train_X_orig, train_Y_orig, test_X_orig, test_Y_orig = load_sin_line_dataset(data_path="../../../data/sin_line.txt")

        print("train_X_orig.shape={}".format(train_X_orig.shape))
        print("train_Y_orig.shape={}".format(train_Y_orig.shape))
        print("test_X_orig.shape={}".format(test_X_orig.shape))
        print("test_Y_orig.shape={}".format(test_Y_orig.shape))

        train_X, train_Y, test_X, test_Y = (train_X_orig, train_Y_orig, test_X_orig, test_Y_orig)
        degree = 20
        train_X = map_feature(train_X, degree)
        test_X = map_feature(test_X, degree)

        print("train_X.shape={}".format(train_X.shape))
        print("train_Y.shape={}".format(train_Y.shape))
        print("test_X.shape={}".format(test_X.shape))
        print("test_Y.shape={}".format(test_Y.shape))

        model = LinearRegression(num_features=train_X.shape[0],
                                    regularizer=L2Regularizer(0), iterator=NumberIterator(100),
                                    optimizer=GradientOptimizer(0.5), initializer=WeightInitializer(),
                                    keep_cost_num_epoch=50)

        model.fit(train_X, train_Y, print_cost=True, print_num_epoch=50)

        print('train cost={}'.format(model.get_cost(train_X, train_Y)))
        print('test cost={}'.format(model.get_cost(test_X, test_Y)))
        print('W={}'.format(model.W))
        print('b={}'.format(model.b))
        predictor = Predictor(model)
        # predictor.plot_regression_curve(train_X_orig, train_Y_orig, X_model=train_X)






import logging
import os
from qbz95.conf.config_base import ConfigBase

logging.root.setLevel(level=logging.INFO)

class ConfigJupyter(ConfigBase):

    def __init__(self):
        ConfigBase.__init__(self, 'jupyter')

    package_path = lambda self: '/notebooks/eipi10/arsenal'
    data_path = lambda self: os.path.join(self.package_path(), 'data')

    http_proxy = lambda self: 'http://web-proxy.rose.hp.com:8080'






from flask import Flask
from redis import Redis

# https://yeasy.gitbooks.io/docker_practice/content/compose/usage.html
# docker-compose up

app = Flask(__name__)
redis = Redis(host='redis', port=6379)

@app.route('/')
def hello():
    count = redis.incr('hits')
    return 'Hello World! the page has been visisted {} times\n'.format(count)

if __name__ == "__main__":
    app.run(host="0.0.0.0", debug=True)





import numpy as np
import pymysql
import psycopg2
import logging
import time
import pandas as pd
import sys
import datetime


from functools import partial
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE


def eval_numerical_gradient(f, x, verbose=False, h=0.00001):
    """Evaluates gradient df/dx via finite differences:
    df/dx ~ (f(x+h) - f(x-h)) / 2h
    Adopted from https://github.com/ddtm/dl-course/ (our ysda course).
    the function is very ingenious
    """
    fx = f(x) # evaluate function value at original point
    grad = np.zeros_like(x)
    # iterate over all indexes in x
    it = np.nditer(x, flags=['multi_index'], op_flags=['readwrite'])
    while not it.finished:

        # evaluate function at x+h
        ix = it.multi_index
        oldval = x[ix]
        x[ix] = oldval + h # increment by h
        fxph = f(x) # evalute f(x + h)
        x[ix] = oldval - h
        fxmh = f(x) # evaluate f(x - h)
        x[ix] = oldval # restore

        # compute the partial derivative with centered formula
        grad[ix] = (fxph - fxmh) / (2 * h) # the slope
        if verbose:
            print (ix, grad[ix])
        it.iternext() # step to next dimension

    return grad


# from preprocessed_mnist import load_dataset
def load_dataset(data_path="../../../data", flatten=False):
    f = np.load(data_path + "/mnist/mnist.npz")
    X_train, y_train = f['x_train'], f['y_train']
    X_test, y_test = f['x_test'], f['y_test']

    # normalize x
    X_train = X_train.astype(float) / 255.
    X_test = X_test.astype(float) / 255.

    # we reserve the last 10000 training examples for validation
    X_train, X_val = X_train[:-10000], X_train[-10000:]
    y_train, y_val = y_train[:-10000], y_train[-10000:]

    if flatten:
        X_train = X_train.reshape([X_train.shape[0], -1])
        X_val = X_val.reshape([X_val.shape[0], -1])
        X_test = X_test.reshape([X_test.shape[0], -1])

    return X_train, y_train, X_val, y_val, X_test, y_test


def iterate_minibatches(X, Y, batch_size, shuffle=False):
    assert len(X) == len(Y)
    if shuffle:
        indices = np.random.permutation(len(X))
    for start_idx in range(0, len(X) - batch_size + 1, batch_size):
        if shuffle:
            excerpt = indices[start_idx:start_idx + batch_size]
        else:
            excerpt = slice(start_idx, start_idx + batch_size)
        yield X[excerpt], Y[excerpt]


def onehot(y, depth=10):
    return np.eye(depth)[y]


def plot_epochs(history, metrics_name):

    train_metrics = history.history[metrics_name]
    val_metrics = history.history['val_'+metrics_name]

    epochs = range(1, len(train_metrics) + 1)

    plt.clf()   # clear figure
    plt.plot(epochs, train_metrics, 'b', label='Training ' + metrics_name, )
    plt.plot(epochs, val_metrics, 'r', label='Validation ' + metrics_name)
    plt.title('Training and validation ' +  metrics_name)
    plt.xlabel('Epochs')
    plt.ylabel(metrics_name)
    plt.legend()

    plt.show()


def plot_tsne(model, perplexity=40, init='pca', n_iter=2500, random_state=23, verbose=0):
    "Creates and TSNE model and plots it"
    labels = []
    tokens = []

    for word in model.wv.vocab:
        tokens.append(model[word])
        labels.append(word)

    tsne_model = TSNE(perplexity=perplexity, n_components=2, init=init, n_iter=n_iter, random_state=random_state, verbose=verbose)
    new_values = tsne_model.fit_transform(tokens)

    x = []
    y = []
    for value in new_values:
        x.append(value[0])
        y.append(value[1])

    plt.figure(figsize=(16, 16))
    for i in range(len(x)):
        plt.scatter(x[i], y[i])
        plt.annotate(labels[i],
                     xy=(x[i], y[i]),
                     xytext=(5, 2),
                     textcoords='offset points',
                     ha='right',
                     va='bottom')
    plt.show()


class ProgressCounter:
    def __init__(self, total_count, step=0, name=''):
        self.total_count = total_count
        self.count = 0
        self.step = step
        self.name = name

    def add_one(self):
        self.count += 1
        current_time = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        if self.step==0 or self.count==self.total_count or self.count % self.step == 0 :
            sys.stdout.write('%s: %s %d/%d\r' % (current_time, self.name, self.count, self.total_count))

class TaskTime:
    def __init__(self, task_name, show_start=False):
        self.show_start = show_start
        self.task_name = task_name
        self.start_time = time.time()

    def elapsed_time(self):
        return time.time()-self.start_time

    def __enter__(self):
        if self.show_start:
            logging.info('start {}'.format(self.task_name))
        return self;

    def __exit__(self, exc_type, exc_value, exc_tb):
        logging.info('finish {} [elapsed time: {:.2f} seconds]'.format(self.task_name, self.elapsed_time()))



class MySqlHelper:

    @classmethod
    def get_connection(cls, host='15.15.165.218', port=3306, user='grid', password='grid' ):
        """
        :return: return the pre-connect database
        """
        connection = pymysql.connect(host=host,
                             user=user,
                             password=password,
                             port=port,
                             charset='utf8', #'utf8mb4'
                             use_unicode=True,)
        return connection


class RedshiftHelper:

    @classmethod
    def get_connection(cls, host='15.15.165.218', port=5439, user='grid', password='grid', dbname='csdw'):
        """
        :return: return the pre-connect database
        """
        logging.info('host={}, user={}, port={}, dbname={}'.format(host, user, port, dbname))
        connection = psycopg2.connect(host=host,
                                      user=user,
                                      password=password,
                                      port=port,
                                      dbname=dbname)
        connection.autocommit = True
        return connection

    @classmethod
    def execute(cls, conn, sqls):
        with conn.cursor() as cursor:
            for sql in sqls:
                print('-' * 100)
                logging.info('execute ' + sql)
                cursor.execute(sql)


    @classmethod
    def execute_file(cls, conn, files):
        with conn.cursor() as cursor:
            for file in files:
                print('-' * 100)
                logging.info('execute ' + file)
                with open(file, 'r') as f:
                    sql = f.read()
                cursor.execute(sql)


class PandasHelper:

    @classmethod
    def read_sql(cls, sql, conn, data_set_name='data'):
        with TaskTime('loading {} from database'.format(data_set_name)):
            df = pd.read_sql(sql, conn)
        logging.info('{} records are loaded'.format(len(df)))
        return df

    @classmethod
    def read_csv(cls, path, header=None, columns=None, parse_dates=None, dtype=None, quoting=0, sep='\t', compression='infer'):
        with TaskTime('loading csv from ' + path):
            if header is None:
                df = pd.read_csv(path, sep=sep, header=None, names=columns, dtype=dtype, quoting=quoting,
                                 parse_dates=parse_dates, compression=compression)
            else:
                df = pd.read_csv(path, sep=sep, header=header, names=None, dtype=dtype, quoting=quoting,
                                 parse_dates=parse_dates, compression=compression)
        logging.info('{} records are loaded'.format(len(df)))
        return df

    @classmethod
    def to_csv(cls, df, path, header=False, index=False, sep='\t', compression=None):
        with TaskTime('saving data frame to {}'.format(path)):
            df.to_csv(path, header=header, index=index, sep=sep, compression=compression)
        logging.info('{} records are saved'.format(len(df)))






# coding=utf-8
import tensorflow as tf
from tensorflow.examples.tutorials.mnist import input_data


# define W & b
def weight_variable(para):
    # ?????????????stddev?0.1
    initial = tf.truncated_normal(para, stddev=0.1)
    return tf.Variable(initial)


def bias_variable(para):
    initial = tf.constant(0.1, shape=para)
    return tf.Variable(initial)




# ??????
def variable_summaries(var):
    with tf.name_scope('summaries'):  # ???scope??
        mean = tf.reduce_mean(var)
        tf.summary.scalar('mean', mean)
        with tf.name_scope('stddev'):
            stddev = tf.sqrt(tf.reduce_mean(tf.reduce_mean(tf.square(var - mean))))
        tf.summary.scalar('stddev', stddev)
        tf.summary.scalar('max', tf.reduce_max(var))
        tf.summary.scalar('min', tf.reduce_min(var))  # ????var? ??????????????
        tf.summary.histogram('histogram', var)  # ????var???? - histogram


# ????? - ?????????
#          - ?????????
#          - ????????relu
def nn_layer(in_data, in_dim, out_dim, layer_name, act=tf.nn.relu):
    with tf.name_scope(layer_name):  # ??????????
        with tf.name_scope('weights'):
            weights = weight_variable([in_dim, out_dim])  # ?????????
            variable_summaries(weights)  # ????????
        with tf.name_scope('biases'):
            biases = bias_variable([out_dim])  # ???????biases
            variable_summaries(biases)  # ??????biases
        with tf.name_scope('Wx_plus_b'):
            Wx_plus_b = tf.matmul(in_data, weights) + biases  # ?? wx?b
            tf.summary.histogram('Wx_plus_b', Wx_plus_b)  # ??wx?b?? - histogram
        activations = act(Wx_plus_b, name='activation')  # ??
        tf.summary.histogram('activations', activations)  # ??????? - histogram
        return activations

    # ?????????session


mnist = input_data.read_data_sets("../../data/mnist/", one_hot=True)
sess = tf.InteractiveSession()

# ????????
with tf.name_scope('input'):
    x = tf.placeholder(tf.float32, [None, 784], name='x_input')  # 28*28=784 dim??????name???????
    y_ = tf.placeholder(tf.float32, [None, 10], name='x_input')  # label - 10 dim

# ????reshape
with tf.name_scope('input_reshape'):
    x_shaped = tf.reshape(x, [-1, 28, 28, 1])  # reshape for conv, -1????????1????
    tf.summary.image('input', x_shaped, 10)  # ??input????? - image

# ?????layer
layer1 = nn_layer(x, 784, 500, 'layer1')

with tf.name_scope('dropout'):  # layer1 ??follow??dropout
    keep_prob = tf.placeholder(tf.float32)
    tf.summary.scalar('dropout_keep_probability', keep_prob)
    dropped = tf.nn.dropout(layer1, keep_prob)

# ?????layer
y = nn_layer(dropped, 500, 10, 'layer2', act=tf.identity)

# softmax
# ?????cross_entropy
with tf.name_scope('cross_entropy'):
    diff = tf.nn.softmax_cross_entropy_with_logits(logits=y, labels=y_)  # diff
    with tf.name_scope('total'):
        cross_entropy = tf.reduce_mean(diff)
tf.summary.scalar('cross_entropy', cross_entropy)



# training
with tf.name_scope('train'):
    train_step = tf.train.AdamOptimizer(0.001).minimize(cross_entropy)
with tf.name_scope('accuracy'):
    with tf.name_scope('correct_prediction'):
        correct_prediction = tf.equal(tf.argmax(y, 1), tf.argmax(y_, 1))
    with tf.name_scope('accuracy'):
        accuracy_op = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))
tf.summary.scalar('accuracy', accuracy_op)

# ????summary
merge_op = tf.summary.merge_all()
train_writer = tf.summary.FileWriter('./logs/mnist/train', sess.graph)  # ??log????
test_writer = tf.summary.FileWriter('./logs/mnist/test', sess.graph)  # ??log????

# ????
tf.global_variables_initializer().run()  # ???????

# ??tf.train.Saver()????????
saver = tf.train.Saver()
for i in range(1000):
    if i % 10 == 0:  # for test??10?step????
        # ????
        xs, ys = mnist.test.images, mnist.test.labels
        # ??merge_op(????) ? accuracy_op(?????)
        summary, acc = sess.run([merge_op, accuracy_op], feed_dict={x: xs, y_: ys, keep_prob: 1.0})
        print('Accuracy at step %s: %s' % (i, acc))  # ??????

    else:  # for train
        if i == 99:
            # tf.RunOption???????
            # tf.RunMetadata()???????????????????????
            run_options = tf.RunOptions(trace_level=tf.RunOptions.FULL_TRACE)
            run_metadata_op = tf.RunMetadata()
            xs, ys = mnist.train.next_batch(100)
            summary, _ = sess.run([merge_op, train_step], feed_dict={x: xs, y_: ys, keep_prob: 0.6},
                                  options=run_options, run_metadata=run_metadata_op)
            train_writer.add_run_metadata(run_metadata_op, 'step % 03d', i)
            train_writer.add_summary(summary, i)
            saver.save(sess, './logs/model.ckpt', i)
            print('Adding run metadata for', i)
        else:
            xs, ys = mnist.train.next_batch(100)
            # ??merged_op?train_step???summary
            summary, _ = sess.run([merge_op, train_step], feed_dict={x: xs, y_: ys, keep_prob: 0.6})
            train_writer.add_summary(summary, i)
        # ???????writer
train_writer.close
test_writer.close


import logging
import os


logging.basicConfig(format='%(asctime)s: %(levelname)s: %(message)s')
logging.root.setLevel(level=logging.WARNING)
logging.logThreads = False
logging.logProcesses = False


class ConfigBase():
    def __init__(self, env):
        self.env = env

    package_path = lambda self: None
    data_path = lambda self: None
    base_path = lambda self: os.path.join(self.package_path(), 'arsenal')

    http_proxy = lambda self: None





import unittest
from qbz95.ml.linear_model import *
from qbz95.ml.utils import *
from qbz95.ml.planar_utils import *
from qbz95.ml.predictor import ClassificationPredictor

# logger.setLevel(level=logging.DEBUG)


class TestLogisticRegression(unittest.TestCase):
    def test_sigmoid(self):
        s = sigmoid(np.array([0, 2]))
        print("sigmoid([0, 2]) = " + str(s))
        np.testing.assert_array_almost_equal(s, np.array([0.5, 0.88079708]))

    def test_initialize_image(self):
        train_set_x_orig, train_set_y_orig, test_set_x_orig, test_set_y_orig, classes = \
            load_image_data(train_data_path='../../../data/train_catvnoncat.h5', test_data_path='../../../data/test_catvnoncat.h5')
        train_set_x, train_set_y = initialize_image_data(train_set_x_orig, train_set_y_orig)
        test_set_x, test_set_y = initialize_image_data(test_set_x_orig, test_set_y_orig)

        print("train_set_x shape: " + str(train_set_x.shape))
        print("train_set_y shape: " + str(train_set_y.shape))
        print("test_set_x shape: " + str(test_set_x.shape))
        print("test_set_y shape: " + str(test_set_y.shape))
        print("sanity check after reshaping: " + str(train_set_x[0:6, 0:4]))
        print("sanity check after reshaping: " + str(train_set_x[0:5, 0] * 255))

        np.testing.assert_array_almost_equal(train_set_x[0:5, 0:4],
                                             np.array([[0.06666667,  0.12156863,  0.21960784,  0.08627451,  0.12941176],
                                                       [0.76862745,  0.75294118,  0.74509804,  0.75686275,  0.72941176],
                                                       [0.32156863,  0.27843137,  0.26666667,  0.34901961,  0.3254902],
                                                       [0.00392157,  0.08627451,  0.00784314,  0.00392157,  0.05490196]]).T)

        np.testing.assert_array_almost_equal(train_set_x[0:5, 0] * 255, [17.,  31.,  56.,  22. , 33.])

    def test_propagate(self):
        # check if model.propagate is correct
        X = np.array([[1., 2., -1.], [3., 4., -3.2]])
        y = np.array([[1], [0], [1]]).T
        W = np.array([[1.], [2.]]).T
        b = 2
        print("X = " + str(X))
        print("y = " + str(y))
        print("W = " + str(W))
        print("b = " + str(b))

        model = LogisticRegression(num_features=X.shape[0], optimizer=GradientOptimizer(0.009),
                                   iterator=NumberIterator(100))
        model.W = W
        model.b = b

        model.propagate(X, y)
        print("dW = " + str(model.dW))
        print("db = " + str(model.db))
        print("cost = " + str(model.cost))

        assert (y.shape == (1, X.shape[1] ))
        np.testing.assert_array_almost_equal(model.dW, np.array([[9.98456015e-01], [2.39507239e+00]]).T)
        np.testing.assert_array_almost_equal(model.db, 1.45557814e-03)
        np.testing.assert_array_almost_equal(model.cost, 5.80154532)

        # check if model.train is correct
        model.fit(X, y, print_cost=True, print_num_epoch=10)

        print("W = " + str(model.W))
        print("b = " + str(model.b))
        print("dW = " + str(model.dW))
        print("db = " + str(model.b))
        np.testing.assert_array_almost_equal(model.W, np.array([[0.19033591], [0.12259159]]).T)
        np.testing.assert_array_almost_equal(model.b, 1.92535983008)
        np.testing.assert_array_almost_equal(model.dW, np.array([[0.67752042], [1.41625495]]).T)
        np.testing.assert_array_almost_equal(model.db, 0.219194504541)



import logging
import os
import six.moves.urllib as urllib


class MnistHelper:
    """提供MNIST手写数字数据集的下载，获取"""

    @classmethod
    def download(cls, target_path, source_url='http://yann.lecun.com/exdb/mnist', http_proxy=None):
        """下载MNIST数据文件"""
        if http_proxy is None:
            proxy_handler = urllib.request.ProxyHandler({'http': http_proxy, 'https': http_proxy})
            opener = urllib.request.build_opener(proxy_handler)
        else:
            opener = urllib.request.build_opener()

        urllib.request.install_opener(opener)

        def maybe_download(file_name):
            if not os.path.exists(target_path):
                os.mkdir(target_path)
            file_path = os.path.join(target_path, file_name)
            if not os.path.exists(file_path):
                source_file_url = os.path.join(source_url, file_name)
                logging.info(source_file_url)
                filepath, _ = urllib.request.urlretrieve(source_file_url, file_path)
                statinfo = os.stat(filepath)
                logging.info('Successfully downloaded {} {} bytes.'.format(file_name, statinfo.st_size))
            else:
                logging.info('Already downloaded {}'.format(file_name))
            return file_path

        train_data_path= maybe_download('train-images-idx3-ubyte.gz')
        train_label_path = maybe_download('train-labels-idx1-ubyte.gz')
        test_data_path= maybe_download('t10k-images-idx3-ubyte.gz')
        test_label_path = maybe_download('t10k-labels-idx1-ubyte.gz')
        return train_data_path, train_label_path, test_data_path, test_label_path


import numpy as np
from qbz95.ml2.optimizer import *
from qbz95.ml2.initializer import *
from qbz95.ml2.regularizer import *


class Layer:
    """
    A building block. Each layer is capable of performing two things:

    - Process input to get output:           output = layer.forward(input)

    - Propagate gradients through itself:    grad_input = layer.backward(input, grad_output)

    Some layers also have learnable parameters which they update during layer.backward.
    """

    def __init__(self):
        pass

    def forward(self, input, train_flag=True):
        return input

    def backward(self, input, grad_output):
        num_units = input.shape[1]
        d_layer_d_input = np.eye(num_units)
        return np.dot(grad_output, d_layer_d_input)  # chain rule


class ReLU(Layer):
    '''Rectified Linear Unit'''
    def __init__(self):
        """ReLU layer simply applies elementwise rectified linear unit to all inputs"""
        Layer.__init__(self)

    def forward(self, input, train_flag=True):
        """Apply elementwise ReLU to [batch, input_units] matrix"""
        return np.maximum(0, input)

    def backward(self, input, grad_output):
        """Compute gradient of loss w.r.t. ReLU input"""
        relu_grad = input > 0
        return grad_output * relu_grad


class Sigmoid(Layer):
    def __init__(self):
        Layer.__init__(self)
        self._output = None

    def forward(self, input, train_flag=True):
        self._output = 1.0 / (1.0 + np.exp(-input))
        return self._output

    def backward(self, input, grad_output):
        if self._output is None: self.forward(input)
        return self._output * (1 - self._output) * grad_output


class Tanh(Layer):
    def __init__(self):
        Layer.__init__(self)
        self._output = None

    def forward(self, input, train_flag=True):
        self._output = np.tanh(input)
        return self._output

    def backward(self, input, grad_output):
        if self._output is None: self.forward(input)
        return (1.0 - np.power(self._output, 2)) * grad_output


class BatchNormalization(Layer):
    def __init__(self, epsilon=1e-8):
        Layer.__init__(self)

        self.epsilon = epsilon
        self.gamma = None
        self.beta = None
        self.t = 0
        self.mu = None
        self.var = None
        self.k = 0
        self.batch_size = 0

        self.initialized = False

    def initial_params(self, var, mu):
        self.gamma = var
        self.beta = mu
        self.initialized = True

    def update_mu_var(self, mu, var, m):
        self.k = self.k + 1
        self.mu = ((self.k - 1)*self.mu + mu)/self.k if self.mu is not None else mu
        self.var = ((self.k - 1)*self.var + m*var/max(1, m-1))/self.k if self.var is not None else m*var/max(1, m-1)

    def compute(self, input):
        mu = np.mean(input, axis=0)
        var = np.mean(np.power(input - mu, 2), axis=0)
        x = (input - mu) / np.sqrt(var + self.epsilon)
        return mu, var, x

    def forward(self, input, train_flag=True):
        # if train_flag:
        if train_flag:
            mu, var, x = self.compute(input)
            self.update_mu_var(mu, var, input.shape[0])
            if not self.initialized:
                self.initial_params(var, mu)
                return input
            else:
               return self.gamma * x + self.beta
        else:
            assert(self.mu is not None and self.var is not None)

            x = (input - self.mu) / np.sqrt(self.var + self.epsilon)
            return self.gamma * x + self.beta

    def backward(self, input, grad_output):
        mu, var, x = self.compute(input)

        if not self.initialized:
            self.initial_params(var, mu)

        grad_gamma = np.sum(grad_output * x, axis=0)
        grad_beta = np.sum(grad_output, axis=0)

        grad_input = self.gamma / np.sqrt(var + self.epsilon) * (grad_output - np.mean(grad_output, axis=0)) - \
                     self.gamma / np.power(var + self.epsilon, 1.5) * (input - mu) * np.mean(grad_output * (input - mu),
                                                                                             axis=0)
        self.gamma = self.gamma - grad_gamma
        self.beta = self.beta - grad_beta

        return grad_input


class MomentumBatchNormalization(BatchNormalization):
    def __init__(self, epsilon=1e-8, decay = 0.99):
        BatchNormalization.__init__(self, epsilon)
        self.decay = decay

    def update_mu_var(self, mu, var, m):
        if self.mu is None: self.mu = np.zeros(mu.shape)
        if self.var is None: self.var = np.zeros(var.shape)
        self.mu = self.decay*self.mu + (1-self.decay)*mu
        self.var = self.decay*self.var + (1-self.decay)*var


class Dropout(Layer):
    def __init__(self, drop_probability=0.1):
        Layer.__init__(self)
        self.drop_probability = drop_probability
        self.keep_probability = 1 - drop_probability

        assert 0 <= self.keep_probability <= 1

        self.mask = None

    def forward(self, input, train_flag=True):
        if train_flag:
            if self.drop_probability == 0:
                return input
            elif self.drop_probability == 1:
                return input.zeros_like()
            else:
                self.mask = np.random.uniform(0, 1.0, input.shape) < self.keep_probability
                return self.mask * input / self.keep_probability
        else:
            return input

    def backward(self, input, grad_output):
        if self.drop_probability == 0:
            return grad_output
        elif self.drop_probability == 1:
            return grad_output.zeros_like()
        else:
            if self.mask is None:
                self.mask = np.random.uniform(0, 1.0, input.shape) < self.keep_probability
            return self.mask * grad_output / self.keep_probability


class Softmax(Layer):
    def __init__(self):
        """
        Softmax

        I think that softmax can be independent from cross entropy
        """
        Layer.__init__(self)
        self._output = None

    def forward(self, input, train_flag=True):
        # e = np.exp(input)
        # s = np.sum(input, axis=-1, keepdims=True)
        # more numerical stability method below
        e = np.exp(input - np.max(input, axis=-1, keepdims=True))
        s = np.sum(e, axis=-1, keepdims=True)
        self._output = e / s
        return self._output

    def backward(self, input, grad_output):
        if self._output is None: self.forward(input)
        x =  self._output * grad_output
        return x - self._output * np.sum(x, axis=-1, keepdims=True)


class Dense(Layer):
    def __init__(self, input_units, output_units, initializer=ScaleInitializer(0.01), regularizer=Regularizer(),
                 optimizer=GradientOptimizer(learning_rate=0.1)):
        """
        A dense layer is a layer which performs a learned affine transformation:
        f(x) = <W*x> + b
        """
        Layer.__init__(self)

        self.weights = initializer.init(input_units, output_units)
        self.biases = np.zeros(output_units)
        self.regularizer = regularizer
        self.optimizer = optimizer

    def forward(self, input, train_flag=True):
        """
        Perform an affine transformation:
        f(x) = <W*x> + b

        input shape: [batch, input_units]
        output shape: [batch, output units]
        """
        return input.dot(self.weights) + self.biases

    def backward(self, input, grad_output):
        # compute d f / d x = d f / d dense * d dense / d x
        # where d dense/ d x = weights transposed
        grad_input = grad_output.dot(self.weights.T)

        # compute gradient w.r.t. weights and biases
        grad_weights = input.T.dot(grad_output) + self.regularizer.gradient(self.weights)  #/ input.shape[0]
        grad_biases = np.sum(grad_output, axis=0) #/ input.shape[0]

        assert grad_weights.shape == self.weights.shape and grad_biases.shape == self.biases.shape
        # Here we perform a stochastic gradient descent step.
        # Later on, you can try replacing that with something better.
        delta_weights, delta_biases = self.optimizer.delta(grad_weights, grad_biases)
        self.weights = self.weights - delta_weights
        self.biases = self.biases - delta_biases

        return grad_input




import unittest
import logging
from qbz95.ml2.utils import *



# logger.setLevel(level=logging.DEBUG)


class TestPandasHelper(unittest.TestCase):
    def test_read_mysql(self):
        sql_message = """select customer_id, session_id, create_time as session_time, dest_type, msg_summary as message from customer_service_prod.up_messages 
        order by customer_id, session_id, create_time limit 20;
        """
        conn = MySqlHelper.get_connection()
        df_message = PandasHelper.read_sql(sql_message, conn, data_set_name='message')
        conn.close()
        # logging.info(len(df_message))

    def test_read_redshift(self):
        sql_message = """select session_id, create_time as session_time, dest_type, msg_summary as message from customer_service_prod.messages_history 
        order by session_id, create_time limit 20;
        """
        conn = RedshiftHelper.get_connection(user='tpch', password='Zeus000!', port=15439)
        df_message = PandasHelper.read_sql(sql_message, conn, data_set_name='message')
        conn.close()
        logging.info(len(df_message))



import unittest
from qbz95.ml.neural_network import *
from qbz95.ml.planar_utils import *
from qbz95.ml.predictor import *
from qbz95.tests.ml.testCases_v2 import *

logger.setLevel(level=logging.DEBUG)


class TestNeuralNetwork(unittest.TestCase):

    def test_layer_size(self):
        X_assess, Y_assess = layer_sizes_test_case()
        model = NeuralNetwork.build_3layers(X_assess, 4, Y_assess)
        print("The size of the input layer is: n_x = " + str(model.layers[0].num_neuron))
        print("The size of the hidden layer is: n_h = " + str(model.layers[1].num_neuron))
        print("The size of the output layer is: n_y = " + str(model.layers[2].num_neuron))
        np.testing.assert_equal(np.array([model.layers[0].num_neuron, model.layers[1].num_neuron, model.layers[2].num_neuron]),
                np.array([5, 4, 2]))

    def test_initialize_weights(self):
        np.random.seed(2)
        model = NeuralNetwork.build_layers([2, 4, 1])

        print("W1 = " + str(model.layers[1].W))
        print("b1 = " + str(model.layers[1].b))
        print("W2 = " + str(model.layers[2].W))
        print("b2 = " + str(model.layers[2].b))

        np.testing.assert_array_almost_equal(
            np.array([[-0.00416758, -0.00056267],
                      [-0.02136196, 0.01640271],
                      [-0.01793436, -0.00841747],
                      [0.00502881, -0.01245288]]),
            model.layers[1].W)

        np.testing.assert_array_almost_equal([[ 0.], [ 0.], [ 0.], [ 0.]], model.layers[1].b)
        np.testing.assert_array_almost_equal([[-0.01057952, -0.00909008,  0.00551454,  0.02292208]], model.layers[2].W)
        np.testing.assert_array_almost_equal([[ 0.]], model.layers[2].b)

    def test_forward_propagation(self):
        X_assess, Y_assess, parameters = forward_propagation_test_case()
        print(X_assess.shape)
        print(Y_assess.shape)
        print(parameters)

        np.random.seed(1)
        model = NeuralNetwork.build_3layers(X_assess, 4, Y_assess)
        model.layers[1].W = parameters['W1']
        model.layers[1].b = parameters['b1']
        model.layers[2].W = parameters['W2']
        model.layers[2].b = parameters['b2']
        model.forward_propagation(X_assess)

        mean = np.array([np.mean(model.layers[1].Z), np.mean(model.layers[1].A),
                         np.mean(model.layers[2].Z), np.mean(model.layers[2].A)])
        print(mean)
        np.testing.assert_array_almost_equal(mean,
                                             [0.262818640198, 0.091999045227, -1.30766601287, 0.212877681719])

    def test_compute_cost(self):
        X_assess, Y_assess, parameters = forward_propagation_test_case()
        model = NeuralNetwork.build_3layers(X_assess, 4, Y_assess)
        A2, Y_assess, parameters = compute_cost_test_case()
        print(A2)
        print(Y_assess)
        print(parameters)

        model.layers[2].W = parameters['W2']
        model.layers[2].b = parameters['b2']
        model.layers[2].A = A2
        cost = model.cost_object.cost(model.layers[2].A, Y_assess)
        print(cost)

        np.testing.assert_array_almost_equal(cost, 0.693058761039)

    def test_backward_propagation(self):
        parameters, cache, X_assess, Y_assess = backward_propagation_test_case()
        print(X_assess.shape)
        print(Y_assess.shape)
        print(parameters)

        np.random.seed(1)
        model = NeuralNetwork.build_3layers(X_assess, 4, Y_assess)
        model.layers[1].W = parameters['W1']
        model.layers[1].b = parameters['b1']
        model.layers[2].W = parameters['W2']
        model.layers[2].b = parameters['b2']
        model.layers[0].A = X_assess
        model.layers[1].Z = cache['Z1']
        model.layers[1].A = cache['A1']
        model.layers[2].Z = cache['Z2']
        model.layers[2].A = cache['A2']
        model.backward_propagation(Y_assess)

        print("dW1 = " + str(model.layers[1].dW))
        print("db1 = " + str(model.layers[1].db))
        print("dW2 = " + str(model.layers[2].dW))
        print("db2 = " + str(model.layers[2].db))
        np.testing.assert_array_almost_equal(model.layers[1].dW, [[0.00301023, -0.00747267],
                                                                  [0.00257968, -0.00641288],
                                                                  [-0.00156892, 0.003893],
                                                                  [-0.00652037, 0.01618243]])
        np.testing.assert_array_almost_equal(model.layers[1].db, [[0.00176201],
                                                                  [0.00150995],
                                                                  [-0.00091736],
                                                                  [-0.00381422]])
        np.testing.assert_array_almost_equal(model.layers[2].dW, [[0.00078841, 0.01765429, -0.00084166, -0.01022527]])
        np.testing.assert_array_almost_equal(model.layers[2].db, [[-0.16655712]])

    def test_fit(self):
        X_assess, Y_assess = nn_model_test_case()
        print(X_assess.shape)
        print(Y_assess.shape)

        np.random.seed(2)
        model = NeuralNetwork.build_3layers(X_assess, 4, Y_assess,
                                            iterator=NumberIterator(10000), learning_rate=1.2)
        model.fit(X_assess, Y_assess, print_cost=True)

        print("W1 = " + str(model.layers[1].W))
        print("b1 = " + str(model.layers[1].b))
        print("W2 = " + str(model.layers[2].W))
        print("b2 = " + str(model.layers[2].b))

        np.testing.assert_array_almost_equal([[-0.65848169, 1.21866811],
                                              [-0.76204273, 1.39377573],
                                              [0.5792005, -1.10397703],
                                              [0.76773391, -1.41477129]], model.layers[1].W)
        np.testing.assert_array_almost_equal([[0.287592], [0.3511264], [-0.2431246], [-0.35772805]], model.layers[1].b)
        np.testing.assert_array_almost_equal([[-2.45566237, -3.27042274 , 2.00784958,  3.36773273]], model.layers[2].W)
        np.testing.assert_array_almost_equal([[ 0.20459656]], model.layers[2].b)


    def test_compute_cost_with_regularization(self):
        A3, Y_assess, parameters = compute_cost_with_regularization_test_case()

        model = NeuralNetwork.build_layers([3, 2, 3, 1], regularizer=L2Regularizer(0.1))
        model.layers[1].W = parameters["W1"]
        model.layers[2].W = parameters["W2"]
        model.layers[3].W = parameters["W3"]
        model.layers[3].A = A3
        model.compute_cost(Y_assess)

        print("cost = " + str(model.cost))
        np.testing.assert_almost_equal(1.78648594516, model.cost)

    def test_backward_propagation_with_regularization(self):
        X_assess, Y_assess, cache = backward_propagation_with_regularization_test_case()
        (Z1, A1, W1, b1, Z2, A2, W2, b2, Z3, A3, W3, b3) = cache


        model = NeuralNetwork.build_layers(
            [X_assess.shape[0], Z1.shape[0], Z2.shape[0], Z3.shape[0]],
            activators=[Relu, Relu, Sigmoid],
            regularizer=L2Regularizer(0.7))
        model.layers[0].A = X_assess
        model.layers[1].W = W1
        model.layers[1].b = b1
        model.layers[1].A = A1
        model.layers[1].Z = Z1
        model.layers[2].W = W2
        model.layers[2].b = b2
        model.layers[2].A = A2
        model.layers[2].Z = Z2
        model.layers[3].W = W3
        model.layers[3].b = b3
        model.layers[3].A = A3
        model.layers[3].Z = Z3

        model.backward_propagation(Y_assess)

        print("dW1 = " + str(model.layers[1].dW))
        print("dW2 = " + str(model.layers[2].dW))
        print("dW3 = " + str(model.layers[3].dW))
        # np.testing.assert_almost_equal(1.78648594516, model.cost)
        np.testing.assert_array_almost_equal([[-0.25604646,  0.12298827, - 0.28297129],
                                              [-0.17706303,  0.34536094, - 0.4410571]], model.layers[1].dW)

        np.testing.assert_array_almost_equal([[0.79276486,  0.85133918],
                                              [-0.0957219, - 0.01720463],
                                              [-0.13100772, - 0.03750433]], model.layers[2].dW)
        np.testing.assert_array_almost_equal([[-1.77691347, - 0.11832879, - 0.09397446]], model.layers[3].dW)

    def test_backward_propagation_with_dropout(self):
        X_assess, parameters = forward_propagation_with_dropout_test_case()
        np.random.seed(1)
        model = NeuralNetwork.build_layers(
            [X_assess.shape[0], 2, 3, 1],
            activators=[Relu, Relu, Sigmoid],
            regularizer=Regularizer(),
            keep_prob = 0.7
        )
        model.layers[1].W = parameters['W1']
        model.layers[1].b = parameters['b1']
        model.layers[2].W = parameters['W2']
        model.layers[2].b = parameters['b2']
        model.layers[3].W = parameters['W3']
        model.layers[3].b = parameters['b3']
        model.layers[0].A = X_assess

        model.forward_propagation(X_assess)
        print("A3 = " + str(model.output_layer.A))
        np.testing.assert_array_almost_equal([[0.36974721, 0.00305176, 0.04565099, 0.49683389, 0.36974721]], model.output_layer.A)

    def test_backward_propagation_with_dropout(self):
        X_assess, Y_assess, cache = backward_propagation_with_dropout_test_case()
        (Z1, D1, A1, W1, b1, Z2, D2, A2, W2, b2, Z3, A3, W3, b3) = cache

        model = NeuralNetwork.build_layers(
            [X_assess.shape[0], Z1.shape[0], Z2.shape[0], Z3.shape[0]],
            activators=[Relu, Relu, Sigmoid],
            regularizer=Regularizer(),
            keep_prob=0.8)
        model.layers[0].A = X_assess
        model.layers[1].W = W1
        model.layers[1].b = b1
        model.layers[1].A = A1
        model.layers[1].D = D1
        model.layers[1].Z = Z1
        model.layers[2].W = W2
        model.layers[2].b = b2
        model.layers[2].A = A2
        model.layers[2].D = D2
        model.layers[2].Z = Z2
        model.layers[3].W = W3
        model.layers[3].b = b3
        model.layers[3].A = A3
        model.layers[3].Z = Z3

        model.backward_propagation(Y_assess)

        print("dZ1 = " + str(model.layers[1].dZ))
        print("dZ2 = " + str(model.layers[2].dZ))
        print("dZ3 = " + str(model.layers[3].dZ))

        np.testing.assert_array_almost_equal([[0.00019884,  0.00028657,  0.00012138],
                                              [0.00035647,  0.00051375,  0.00021761]], model.layers[1].dW)

        np.testing.assert_array_almost_equal([[-0.00256518, - 0.0009476],
                                              [0.,        0.],
                                              [0.,        0.]], model.layers[2].dW)
        np.testing.assert_array_almost_equal([[-0.06951191,  0.,         0.]], model.layers[3].dW)


    def test_gradient_check(self):
        X, Y, parameters = gradient_check_n_test_case()

        model = NeuralNetwork.build_layers(
            [X.shape[0], 5, 3, Y.shape[0]],
            activators=[Relu, Relu, Sigmoid],
            regularizer=L2Regularizer(10),
            keep_prob=0.65)

        model.layers[1].W = parameters['W1']
        model.layers[1].b = parameters['b1']
        model.layers[2].W = parameters['W2']
        model.layers[2].b = parameters['b2']
        model.layers[3].W = parameters['W3']
        model.layers[3].b = parameters['b3']
        model.layers[0].A = X

        return model.gradient_check(X, Y)

    def test_adam(self):
        parameters, grads, v, s = update_parameters_with_adam_test_case()

        model = NeuralNetwork.build(
            [3, 2, 3],
            activators=[Relu, Sigmoid],
            optimizer=AdamOptimizer(learning_rate = 0.01, beta1 = 0.9, beta2 = 0.999,  epsilon = 1e-8))
        model.layers[1].W = parameters["W1"]
        model.layers[1].b = parameters["b1"]
        model.layers[1].t = 1
        model.layers[2].W = parameters["W2"]
        model.layers[2].b = parameters["b2"]
        model.layers[2].t = 1

        model.layers[1].dW = grads["dW1"]
        model.layers[1].db = grads["db1"]
        model.layers[2].dW = grads["dW2"]
        model.layers[2].db = grads["db2"]

        model.update_parameters()

        print("W1 = " + str(model.layers[1].W))
        print("b1 = " + str(model.layers[1].b))
        print("W2 = " + str(model.layers[2].W))
        print("b2 = " + str(model.layers[2].b))
        print("v[\"dW1\"] = " + str(model.layers[1].vW))
        print("v[\"db1\"] = " + str(model.layers[1].vb))
        print("v[\"dW2\"] = " + str(model.layers[2].vW))
        print("v[\"db2\"] = " + str(model.layers[2].vb))
        print("s[\"dW1\"] = " + str(model.layers[1].sW))
        print("s[\"db1\"] = " + str(model.layers[1].sb))
        print("s[\"dW2\"] = " + str(model.layers[2].sW))
        print("s[\"db2\"] = " + str(model.layers[2].sb))

        np.testing.assert_array_almost_equal([[ 1.63178673, -0.61919778, -0.53561312],
                                              [-1.08040999, 0.85796626, -2.29409733]], model.layers[1].W)
        np.testing.assert_array_almost_equal([[ 1.75225313], [-0.75376553]], model.layers[1].b)
        np.testing.assert_array_almost_equal([[ 0.32648046, -0.25681174, 1.46954931],
                                              [-2.05269934, -0.31497584, -0.37661299],
                                              [ 1.14121081, -1.09245036, -0.16498684]], model.layers[2].W)
        np.testing.assert_array_almost_equal([[-0.88529978], [ 0.03477238], [ 0.57537385]], model.layers[2].b)
        np.testing.assert_array_almost_equal([[-0.11006192, 0.11447237, 0.09015907],
                                              [ 0.05024943, 0.09008559, -0.06837279]], model.layers[1].vW)
        np.testing.assert_array_almost_equal([[-0.01228902], [-0.09357694]], model.layers[1].vb)
        np.testing.assert_array_almost_equal([[-0.02678881, 0.05303555, -0.06916608],
                                              [-0.03967535, -0.06871727, -0.08452056],
                                              [-0.06712461, -0.00126646, -0.11173103]], model.layers[2].vW)
        np.testing.assert_array_almost_equal([[ 0.02344157], [ 0.16598022], [ 0.07420442]], model.layers[2].vb)
        np.testing.assert_array_almost_equal([[ 0.00121136, 0.00131039, 0.00081287], [ 0.0002525, 0.00081154, 0.00046748]], model.layers[1].sW)
        np.testing.assert_array_almost_equal([[ 1.51020075e-05], [ 8.75664434e-04]], model.layers[1].sb)
        np.testing.assert_array_almost_equal([[ 7.17640232e-05, 2.81276921e-04, 4.78394595e-04],
                                              [ 1.57413361e-04, 4.72206320e-04, 7.14372576e-04],
                                              [ 4.50571368e-04, 1.60392066e-07, 1.24838242e-03]], model.layers[2].sW)
        np.testing.assert_array_almost_equal([[ 5.49507194e-05], [ 2.75494327e-03], [ 5.50629536e-04]], model.layers[2].sb)

    def test_momentum(self):
        parameters, grads, v = update_parameters_with_momentum_test_case()

        model = NeuralNetwork.build(
            [3, 2, 3],
            activators=[Relu, Sigmoid],
            optimizer=MomentumOptimizer(learning_rate = 0.01, beta = 0.9))
        model.layers[1].W = parameters["W1"]
        model.layers[1].b = parameters["b1"]
        model.layers[2].W = parameters["W2"]
        model.layers[2].b = parameters["b2"]

        model.layers[1].dW = grads["dW1"]
        model.layers[1].db = grads["db1"]
        model.layers[2].dW = grads["dW2"]
        model.layers[2].db = grads["db2"]

        model.update_parameters()

        print("W1 = " + str(model.layers[1].W))
        print("b1 = " + str(model.layers[1].b))
        print("W2 = " + str(model.layers[2].W))
        print("b2 = " + str(model.layers[2].b))
        print("v[\"dW1\"] = " + str(model.layers[1].vW))
        print("v[\"db1\"] = " + str(model.layers[1].vb))
        print("v[\"dW2\"] = " + str(model.layers[2].vW))
        print("v[\"db2\"] = " + str(model.layers[2].vb))

        np.testing.assert_array_almost_equal([[ 1.62544598, -0.61290114, -0.52907334],
                                              [-1.07347112, 0.86450677, -2.30085497]], model.layers[1].W)
        np.testing.assert_array_almost_equal([[ 1.74493465], [-0.76027113]], model.layers[1].b)
        np.testing.assert_array_almost_equal([[ 0.31930698, -0.24990073, 1.4627996 ],
                                              [-2.05974396, -0.32173003, -0.38320915],
                                              [ 1.13444069, -1.0998786, -0.1713109 ]], model.layers[2].W)
        np.testing.assert_array_almost_equal([[-0.87809283], [ 0.04055394], [ 0.58207317]], model.layers[2].b)
        np.testing.assert_array_almost_equal([[-0.11006192, 0.11447237, 0.09015907],
                                              [ 0.05024943, 0.09008559, -0.06837279]], model.layers[1].vW)
        np.testing.assert_array_almost_equal([[-0.01228902], [-0.09357694]], model.layers[1].vb)
        np.testing.assert_array_almost_equal([[-0.02678881, 0.05303555, -0.06916608],
                                              [-0.03967535, -0.06871727, -0.08452056],
                                              [-0.06712461, -0.00126646, -0.11173103]], model.layers[2].vW)
        np.testing.assert_array_almost_equal([[ 0.02344157], [ 0.16598022], [ 0.07420442]], model.layers[2].vb)


    def test_interate(self):
        train_X_orig, train_Y_orig, test_X_orig, test_Y_orig = load_petal_dataset()
        print("train_X_orig.shape={}".format(train_X_orig.shape))
        print("train_Y_orig.shape={}".format(train_Y_orig.shape))
        print("test_X_orig.shape={}".format(test_X_orig.shape))
        print("test_Y_orig.shape={}".format(test_Y_orig.shape))

        learning_rate = 0.025

        # batch
        epoch_times = 1000
        mini_batch = 100
        print_num_epoch = 50
        keep_cost_num_epoch = 50

        # initialize
        static_weight_scale = 10

        # regularize
        alpha = 0.01
        keep_prob = 1

        # optimize
        beta = 0.9
        beta1 = 0.9
        beta2 = 0.99
        epsilon = 1e-8

        train_X, train_Y, test_X, test_Y = (train_X_orig, train_Y_orig, test_X_orig, test_Y_orig)
        # train_X, train_Y, test_X, test_Y = load_petal_dataset(num_example=2000) #?????????????
        num_neurons = [train_X.shape[0], 16, 8, 1]
        activators = [Relu, Relu, Sigmoid]

        model = NeuralNetwork.build(
            num_neurons=num_neurons,
            activators=activators,
            iterator=NumberIterator(epoch_times),
            initializer=HeWeightInitializer(),
            optimizer=AdamOptimizer(learning_rate=learning_rate, beta1=beta1, beta2=beta2, epsilon=epsilon),
            keep_cost_num_epoch=keep_cost_num_epoch,
            mini_batch=mini_batch,
            regularizer=L2Regularizer(alpha),
            keep_prob=keep_prob
        )

        model.fit(train_X, train_Y, print_cost=False, print_num_epoch=print_num_epoch)
        # model.plot_costs()

        predictor = ClassificationPredictor(model=model)
        predictor.print_metrics(train_X, train_Y, test_X=test_X, test_Y=test_Y)

        predictor = ProbabilityPredictor(model=model)
        predictor.print_metrics(train_X, train_Y, test_X=test_X, test_Y=test_Y)




import tensorflow as tf
import keras
from keras.layers import Conv1D, MaxPooling1D, Embedding
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences


embedding_layer = Embedding(num_words,
                            EMBEDDING_DIM,
                            embeddings_initializer=Constant(embedding_matrix),
                            input_length=MAX_SEQUENCE_LENGTH,
                            trainable=False)

tokenizer = Tokenizer(num_words=MAX_NUM_WORDS)
from sklearn.model_selection import train_test_split
from keras.preprocessing.sequence import pad_sequences


import pandas as pd
df = pd.read_csv()
df.mean

from scipy import optimize
optimize.fmin_slsqp
from scipy.optimize import minimize



import numpy as np


class Bandit:
    def __init__(self, mean, mean_return=0):
        self.mean = mean
        self.mean_return = mean_return
        self.n = 0

    def pull(self):
        ret = np.random.randn() + self.mean
        return ret

    def update(self, ret):
        self.n += 1
        self.mean_return = (1 - 1.0 / self.n) * self.mean_return + 1.0 / self.n * ret

    def pull_update(self):
        ret = self.pull()
        self.update(ret)
        return ret


class BonuliBandit(Bandit):
    def __init__(self, mean, mean_return=0):
        assert(mean <= 1)
        Bandit.__init__(self, mean, mean_return)
        self.win = 0
        self.lost = 0

    def pull(self):
        if np.random.randn() <= self.mean:
            return 1
        else:
            return 0

    def update(self, ret):
        self.n += 1
        self.mean_return = (1 - 1.0 / self.n) * self.mean_return + 1.0 / self.n * ret
        if ret > 0:
            self.win += 1
        else:
            self.lost += 1





# -*- coding: utf-8 -*-
from chatterbot import ChatBot

# Uncomment the following lines to enable verbose logging
# import logging
# logging.basicConfig(level=logging.INFO)

# Create a new ChatBot instance
bot = ChatBot(
    'Hello',
    storage_adapter='chatterbot.storage.MongoDatabaseAdapter',
    logic_adapters=[
        'chatterbot.logic.BestMatch'
    ],
    filters=[
        'chatterbot.filters.RepetitiveResponseFilter'
    ],
    input_adapter='chatterbot.input.TerminalAdapter',
    output_adapter='chatterbot.output.TerminalAdapter',
    database='chatterbot-database',
    database_uri='mongodb://172.17.0.1:27017',
)

print('Type something to begin...')

while True:
    try:
        bot_input = bot.get_response(None)
        print("-------------------------------")

    # Press ctrl-c or ctrl-d on the keyboard to exit
    except (KeyboardInterrupt, EOFError, SystemExit):
        break


import insuranceqa_data as insuranceqa
import os

# train_data = insuranceqa.load_pairs_train()
test_data = insuranceqa.load_pairs_test()
# valid_data = insuranceqa.load_pairs_valid()

train_data = insuranceqa.load_pool_train()
print(train_data)

vocab_data = insuranceqa.load_pairs_vocab()
print(vocab_data['word2id']['UNKNOWN'])
word = vocab_data['id2word']['10']
print(word)
print(vocab_data['tf'][word])
print(vocab_data['total'])
print(test_data[0])
print(test_data[1])


def ids_to_sentence(ids, id2word):
    words = [id2word[str(id)] for id in ids]
    sentence = ''.join(words)
    return sentence


for i in range(20):
    x = test_data[i]
    question_ids = x['question']
    utterance_ids = x['utterance']
    question = ids_to_sentence(question_ids, vocab_data['id2word'])
    utterance = ids_to_sentence(utterance_ids, vocab_data['id2word'])

    # print(x)
    print('question_ids: %s \nquestion: %s \nutterance: %s' % (question_ids, question, utterance))
    print('-------------------------------------------------')

# valid_data, test_data and train_data share the same properties
# for x in test_data:
#     print('index %s value: %s ++$++ %s ++$++ %s' % \
#      (x['qid'], x['question'], x['utterance'], x['label']))

# vocab_data = insuranceqa.load_pairs_vocab()
# vocab_data['word2id']['UNKNOWN']
# vocab_data['id2word'][0]
# vocab_data['tf']
# vocab_data['total']


# -*- coding: utf-8 -*-
"""
Created on Tue Oct  7 15:43:41 2014

@author: liuweizhi
https://github.com/Greenwicher/Awsome-Project/blob/master/IE5504%20Systems%20Modeling%20and%20Advanced%20Simulation/code/mab.py
"""
import os
import random

random.seed(5504)
import math
import matplotlib.pyplot as plt
from matplotlib import rc

rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica']})
rc('text', usetex=True)
import numpy as np
from scipy.stats import norm
import time


class BernoulliArm():
    def __init__(self, p):
        # generate the probability of success for the arm
        self.p = p
        return

    def draw(self):
        if random.random() > self.p:
            return 0.0
        else:
            return 1.0


class UCB():
    def __init__(self, cp):
        # generate the weight of exploration for the UCB value
        self.cp = cp
        self.counts = []
        self.values = []
        self.ucb_values = []
        return

    def initialize(self, n_arms, arms, init_sim):
        self.counts = [0 for i in range(n_arms)]
        self.values = [0.0 for i in range(n_arms)]
        self.ucb_values = [0.0 for i in range(n_arms)]
        return

    def select_arm(self):
        n_arms = len(self.counts)
        # obtain the initial information for each arm
        for arm in range(n_arms):
            if self.counts[arm] == 0:
                return arm
        # use UCB value to select arm after the initial information obtaining procedure
        self.ucb_values = [0.0 for arm in range(n_arms)]
        total_counts = sum(self.counts)
        for arm in range(n_arms):
            exploration = math.sqrt((2 * math.log(total_counts)) / (float(self.counts[arm])))
            self.ucb_values[arm] = round(self.values[arm] + self.cp * exploration, 5)
        return ind_max(self.ucb_values)

    def update(self, choose_arm, reward):
        self.counts[choose_arm] = self.counts[choose_arm] + 1
        n = self.counts[choose_arm]
        value = self.values[choose_arm]
        new_value = (value * (n - 1) + reward) / (float(n))
        self.values[choose_arm] = new_value
        return


class BayesBandit():
    def __init__(self):
        self.counts = []
        # store the success times for each arm
        self.success = []
        self.bayes = []
        return

    def initialize(self, n_arms, arms, init_sim):
        self.counts = [0 for i in range(n_arms)]
        self.success = [0 for i in range(n_arms)]
        self.bayes = [0 for i in range(n_arms)]
        return

    def select_arm(self):
        for i in range(len(self.bayes)):
            self.bayes[i] = np.random.beta(1 + self.success[i], 1 + self.counts[i] - self.success[i])
        return ind_max(self.bayes)

    def update(self, choose_arm, reward):
        self.counts[choose_arm] += 1
        self.success[choose_arm] += reward
        return


class OCBA():
    def __init__(self):
        self.success = []  # record reward history for each arm
        self.mu = []  # record average reward for each arm
        self.std = []  # record standard deviation for each arm
        self.sn_ratio = []  # record the signal to noise ratio for each arm
        self.ratio = []  # record the allocation ratio for each arm
        self.last_count = []  # record the cumulative counts for each arm in last time
        self.new_count = []  # record the expected counts for each arm in present state based on OCBA allocation rule
        self.n_arms = -9999
        self.init_sim = -9999
        self.arms = []  # store the arms
        self.small_num = 1e-100
        return

    def initialize(self, n_arms, arms, init_sim):
        self.n_arms = n_arms
        # store the reward history for each arm to compute mean and variance
        self.success = [[] for i in range(n_arms)]
        self.mu = [0 for i in range(n_arms)]
        self.std = [0 for i in range(n_arms)]
        # store the signal-to-noise ratio
        self.sn_ratio = [0 for i in range(n_arms)]
        # store the budget allocation ratio
        self.ratio = [0 for i in range(n_arms)]
        # store the trial number for each arm in up to last time
        self.last_count = [0 for i in range(n_arms)]
        # store the optimal trial number for each arm up to now
        self.new_count = [0 for i in range(n_arms)]
        # store the initial simulation time for each arm
        self.init_sim = init_sim
        self.arms = [arm for arm in arms]

        # run initial simulation for each arm with init_sim times to obtain preliminary statistics
        for arm in range(n_arms):
            for foo in range(self.init_sim):
                self.success[arm].append(self.arms[arm].draw())
                self.last_count[arm] += 1
        # compute the preliminary statistic for each arm
        for arm in range(n_arms):
            num_trial = float(len(self.success[arm]))
            self.mu[arm] = np.mean(np.array(self.success[arm]))
            self.std[arm] = math.sqrt(num_trial / (num_trial - 1)) * np.std(np.array(self.success[arm]))
        # find the current best arm
        best_arm = ind_max(self.mu)
        # compute the signal to noise ratio based on OCBA allocation rule
        best_arm_factor = 0  # compute the factor in the signal to noise ratio for best arm (that sqrt{} part)
        for arm in range(n_arms):
            # the signal to noise ratio is no sense for best arm
            if arm == best_arm:
                continue
            else:
                # compute the signal noise ratio for non-best arm
                self.sn_ratio[arm] = (self.std[arm] / (self.mu[best_arm] - self.mu[arm] + self.small_num)) ** 2
                best_arm_factor += (self.sn_ratio[arm] / (self.std[arm] + self.small_num)) ** 2
        # compute the "signal to noise ratio" for the best arm
        self.sn_ratio[best_arm] = self.std[best_arm] * math.sqrt(best_arm_factor)
        # compute the allocation ratio and new count for each arm
        sum_sn_ratio = np.sum(np.array(self.sn_ratio))
        sum_last_count = np.sum(np.array(self.last_count))
        for arm in range(n_arms):
            self.ratio[arm] = self.sn_ratio[arm] / sum_sn_ratio
            self.new_count[arm] = self.ratio[arm] * (sum_last_count + 1)  # delta = 1, in order to avoid numerical issue
        return

    def select_arm(self):
        return ind_max(list(np.array(self.new_count) - np.array(self.last_count)))
        # return ind_max(self.mu)

    def update(self, choose_arm, reward):
        # update basic infor for choose_arm
        self.success[choose_arm].append(reward)
        self.last_count[choose_arm] += 1

        # update basic statistics
        for arm in range(n_arms):
            num_trial = float(len(self.success[arm]))
            self.mu[arm] = np.mean(np.array(self.success[arm]))
            self.std[arm] = math.sqrt(num_trial / (num_trial - 1)) * np.std(
                np.array(self.success[arm])) + self.small_num
        # find the current best arm
        best_arm = ind_max(self.mu)
        # compute the signal to noise ratio based on OCBA allocation rule
        best_arm_factor = 0  # compute the factor in the signal to noise ratio for best arm (that sqrt{} part)
        for arm in range(n_arms):
            # the signal to noise ratio is no sense for best arm
            if arm == best_arm:
                continue
            else:
                # compute the signal noise ratio for non-best arm
                self.sn_ratio[arm] = (self.std[arm] / (self.mu[best_arm] - self.mu[arm] + self.small_num)) ** 2
                best_arm_factor += (self.sn_ratio[arm] / (self.std[arm])) ** 2
        # compute the "signal to noise ratio" for the best arm
        self.sn_ratio[best_arm] = self.std[best_arm] * math.sqrt(best_arm_factor)
        # compute the allocation ratio and new count for each arm
        sum_sn_ratio = np.sum(np.array(self.sn_ratio))
        sum_last_count = np.sum(np.array(self.last_count))
        for arm in range(n_arms):
            self.ratio[arm] = self.sn_ratio[arm] / sum_sn_ratio
            self.new_count[arm] = self.ratio[arm] * (sum_last_count + 1)  # delta = 1, in order to avoid numerical issue
        return


# return the identifier of the largest value in the list
def ind_max(foo):
    return foo.index(max(foo))


# calculate the APCS-B for the current situation
def pcs_calc(trials, reward_history):
    n_arms = len(trials)
    pcs_value = 0
    # compute the estimated success rate for each arm
    arm_mu = [0 for i in range(n_arms)]
    arm_var = [0 for i in range(n_arms)]
    for arm in range(n_arms):
        # to calculate mean and variance, the arm should be played at least once
        if reward_history[arm]:
            arm_mu[arm] = np.mean(np.array(reward_history[arm]))
            arm_var[arm] = np.var(np.array(reward_history[arm]))
    # compute the APCS-B
    best_arm = ind_max(arm_mu)
    # print 'best_arm = %d' % (best_arm)
    for arm in range(n_arms):
        if (arm != best_arm):
            if (trials[arm] >= 2) and (trials[best_arm] >= 2):
                pcs_value += norm.cdf((arm_mu[arm] - arm_mu[best_arm]) / math.sqrt(
                    (arm_var[best_arm] / trials[best_arm]) + (arm_var[arm] / trials[arm]) + 1e-100))
            # print 'Best (%f, %f) , Other (%f, %f), pcs_part = %f' % (arm_mu[best_arm], arm_var[best_arm], arm_mu[arm], arm_var[arm],
            #                                                      norm.cdf((arm_mu[arm] - arm_mu[best_arm]) / math.sqrt(arm_var[best_arm] / trials[best_arm] + arm_var[arm] / trials[arm])))
            else:
                return 0
    if pcs_value == 0.5:
        print
        'PCS = %f' % (1 - pcs_value)
        print
        'Reawrd - History', reward_history
        print
        'len reward history = ', len(reward_history)
        print
        'arm_mu = ', arm_mu
        print
        'arm_var/trials = ', np.array(arm_var) / np.array(trials)
    return 1 - pcs_value


# run simulation and test the performance of algorithims
def run_algorithm(algo, algo_name, arms, n_reps, n_budgets, init_sim):
    # run independent simulation for n_reps time
    rewards = [[0 for i in range(n_budgets)] for j in range(n_reps)]
    cpu_time = [0 for i in range(n_reps)]
    cumulative_rewards = [[0 for i in range(n_budgets)] for j in range(n_reps)]
    pcs = [[0 for i in range(n_budgets)] for j in range(n_reps)]

    for rep in range(n_reps):
        # initialize the algorithm and conduct initial simulation
        n_arms = len(arms)
        algo.initialize(n_arms, arms, init_sim)
        start = n_arms * init_sim
        trials = [0 for i in range(n_arms)]
        reward_history = [[] for i in range(n_arms)]
        time_start = time.time()
        if algo_name == 'OCBA':
            trials = [n_arms * init_sim for i in range(n_arms)]
            reward_history = [[algo.success[arm][step] for step in range(init_sim)] for arm in range(n_arms)]
            rewards[rep][0:start] = [foo for arm in range(n_arms) for foo in reward_history[arm]]
            for i in range(len(rewards[rep])):
                if i == 0:
                    cumulative_rewards[rep][i] = rewards[rep][i]
                else:
                    cumulative_rewards[rep][i] = cumulative_rewards[rep][i - 1] + rewards[rep][i]
        # conduct normal simulation
        for sim in range(start, n_budgets):
            # run algorithm
            choose_arm = algo.select_arm()
            reward = arms[choose_arm].draw()
            algo.update(choose_arm, reward)

            # update statistics
            rewards[rep][sim] = reward
            try:
                cumulative_rewards[rep][sim] = cumulative_rewards[rep][sim - 1] + reward
            except:
                cumulative_rewards[rep][sim] = reward
            trials[choose_arm] += 1
            reward_history[choose_arm].append(reward)
            pcs[rep][sim] = pcs_calc(trials, reward_history)
            if pcs[rep][sim] == 0.5:
                print
                choose_arm
                # print '%s - (%d, %d) : pcs = %f, cumulative reward = %d, choose arm = %d' % (algo_name, rep, sim, pcs[rep][sim], cumulative_rewards[rep][sim], choose_arm)
        time_end = time.time()
        cpu_time[rep] = time_end - time_start

    return [rewards, cpu_time, cumulative_rewards, pcs]


def plot(n_arms, n_budgets, n_reps, rewards, cpu_time, cumulative_rewards, pcs):
    img_dir = os.getcwd() + '/imgs'
    algo_list = [k for k in rewards]

    # boxplot for rewards
    ensemble_rewards = {}
    for algo in algo_list:
        ensemble_rewards[algo] = np.array([0 for i in range(n_budgets)])
        for rep in range(n_reps):
            ensemble_rewards[algo] += np.array(rewards[algo][rep])
        ensemble_rewards[algo] = list(ensemble_rewards[algo] / float(n_reps))
    ylabel = 'Reward'
    title = 'Reward Comparision \n (Replication = %d, Budget = %d, Arms = %d)' % (n_reps, n_budgets, n_arms)
    save_dir = os.path.join(img_dir, 'boxplot_rewards.pdf')
    boxplot(ensemble_rewards, algo_list, ylabel, title, save_dir)

    # boxplot for cpu_time
    ylabel = 'CPU Time (seconds)'
    title = 'CPU Time Comparision \n (Replication = %d, Budget = %d, Arms = %d)' % (n_reps, n_budgets, n_arms)
    save_dir = os.path.join(img_dir, 'boxplot_cpu_time.pdf')
    boxplot(cpu_time, algo_list, ylabel, title, save_dir)

    # time-series plot for pcs
    ensemble_pcs = {}
    for algo in algo_list:
        ensemble_pcs[algo] = [0 for i in range(n_budgets)]
        for rep in range(n_reps):
            ensemble_pcs[algo] += np.array(pcs[algo][rep])
        ensemble_pcs[algo] = list(ensemble_pcs[algo] / float(n_reps))
        plt.plot([i for i in range(n_budgets)], ensemble_pcs[algo], label=algo)
    plt.xlabel('Budgets')
    plt.ylabel('APCS-B')
    plt.title('APCS-B vs Budgets \n (Replication = %d, Arms = %d)' % (n_reps, n_arms))
    box = plt.gca().get_position()
    plt.gca().set_position([box.x0, box.y0 + box.height * 0.2, box.width, box.height * 0.8])
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), fancybox=True, shadow=True, ncol=len(algo_list))
    plt.grid()
    plt.savefig(os.path.join(img_dir, 'series_pcs.pdf'))
    plt.close()

    # time-series plot for cumulative rewards
    ensemble_cum_rewards = {}
    for algo in algo_list:
        ensemble_cum_rewards[algo] = [0 for i in range(n_budgets)]
        for rep in range(n_reps):
            ensemble_cum_rewards[algo] += np.array(cumulative_rewards[algo][rep])
        ensemble_cum_rewards[algo] = list(ensemble_cum_rewards[algo] / float(n_reps))
        plt.plot([i for i in range(n_budgets)], ensemble_cum_rewards[algo], label=algo)
    plt.xlabel('Budgets')
    plt.ylabel('Cumulative Rewards')
    plt.title('Cumulative Rewards vs Budgets \n (Replication = %d, Arms = %d)' % (n_reps, n_arms))
    box = plt.gca().get_position()
    plt.gca().set_position([box.x0, box.y0 + box.height * 0.2, box.width, box.height * 0.8])
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), fancybox=True, shadow=True, ncol=len(algo_list))
    plt.grid()
    plt.savefig(os.path.join(img_dir, 'series_cumulative_rewards.pdf'))
    plt.close()

    return


def boxplot(d, algo_list, ylabel, title, save_dir):
    data = []
    for foo in algo_list:
        data.append(d[foo])
    plt.boxplot(data)
    plt.xticks([i + 1 for i in range(len(algo_list))], tuple(algo_list), rotation=15)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.savefig(save_dir)
    plt.close()

    return


def plot2(algo_name, algo_list, n_arms, n_budgets, n_reps, rewards, cpu_time, cumulative_rewards, pcs):
    img_dir = os.getcwd() + '/imgs'

    # time-series plot for pcs
    ensemble_pcs = {}
    for algo in algo_list:
        ensemble_pcs[algo] = [0 for i in range(n_budgets)]
        for rep in range(n_reps):
            ensemble_pcs[algo] += np.array(pcs[algo][rep])
        ensemble_pcs[algo] = list(ensemble_pcs[algo] / float(n_reps))
        plt.plot([i for i in range(n_budgets)], ensemble_pcs[algo], label=algo)
    plt.xlabel('Budgets')
    plt.ylabel('APCS-B')
    plt.title('APCS-B vs Budgets - %s \n (Replication = %d, Arms = %d)' % (algo_name, n_reps, n_arms))
    box = plt.gca().get_position()
    plt.gca().set_position([box.x0, box.y0 + box.height * 0.2, box.width, box.height * 0.8])
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), fancybox=True, shadow=True, ncol=len(algo_list))
    plt.grid()
    plt.savefig(os.path.join(img_dir, '%s_rate_series_pcs.pdf' % (algo_name)))
    plt.close()

    # time-series plot for cumulative rewards
    ensemble_cum_rewards = {}
    for algo in algo_list:
        ensemble_cum_rewards[algo] = [0 for i in range(n_budgets)]
        for rep in range(n_reps):
            ensemble_cum_rewards[algo] += np.array(cumulative_rewards[algo][rep])
        ensemble_cum_rewards[algo] = list(ensemble_cum_rewards[algo] / float(n_reps))
        plt.plot([i for i in range(n_budgets)], ensemble_cum_rewards[algo], label=algo)
    plt.xlabel('Budgets')
    plt.ylabel('Cumulative Rewards')
    plt.title('Cumulative Rewards vs Budgets - %s \n (Replication = %d, Arms = %d)' % (algo_name, n_reps, n_arms))
    box = plt.gca().get_position()
    plt.gca().set_position([box.x0, box.y0 + box.height * 0.2, box.width, box.height * 0.8])
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), fancybox=True, shadow=True, ncol=len(algo_list))
    plt.grid()
    plt.savefig(os.path.join(img_dir, '%s_rate_series_cumulative_rewards.pdf' % (algo_name)))
    plt.close()

    return


def plot3(algo_list, n_arms, n_budgets, n_reps, rewards, cpu_time, cumulative_rewards, pcs):
    img_dir = os.getcwd() + '/imgs'
    algo_name = 'OCBA'
    # time-series plot for pcs
    ensemble_pcs = {}
    for algo in algo_list:
        ensemble_pcs[algo] = [0 for i in range(n_budgets)]
        for rep in range(n_reps):
            ensemble_pcs[algo] += np.array(pcs[algo][rep])
        ensemble_pcs[algo] = list(ensemble_pcs[algo] / float(n_reps))
        plt.plot([i for i in range(n_budgets)], ensemble_pcs[algo], label=algo)
    plt.xlabel('Budgets')
    plt.ylabel('APCS-B')
    plt.title('APCS-B vs Budgets - %s \n (Replication = %d, Arms = %d)' % (algo_name, n_reps, n_arms))
    box = plt.gca().get_position()
    plt.gca().set_position([box.x0, box.y0 + box.height * 0.2, box.width, box.height * 0.8])
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), fancybox=True, shadow=True, ncol=len(algo_list))
    plt.grid()
    plt.savefig(os.path.join(img_dir, '%s_initsim_series_pcs.pdf' % (algo_name)))
    plt.close()

    # time-series plot for cumulative rewards
    ensemble_cum_rewards = {}
    for algo in algo_list:
        ensemble_cum_rewards[algo] = [0 for i in range(n_budgets)]
        for rep in range(n_reps):
            ensemble_cum_rewards[algo] += np.array(cumulative_rewards[algo][rep])
        ensemble_cum_rewards[algo] = list(ensemble_cum_rewards[algo] / float(n_reps))
        plt.plot([i for i in range(n_budgets)], ensemble_cum_rewards[algo], label=algo)
    plt.xlabel('Budgets')
    plt.ylabel('Cumulative Rewards')
    plt.title('Cumulative Rewards vs Budgets - %s \n (Replication = %d, Arms = %d)' % (algo_name, n_reps, n_arms))
    box = plt.gca().get_position()
    plt.gca().set_position([box.x0, box.y0 + box.height * 0.2, box.width, box.height * 0.8])
    plt.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), fancybox=True, shadow=True, ncol=len(algo_list))
    plt.grid()
    plt.savefig(os.path.join(img_dir, '%s_initsim_series_cumulative_rewards.pdf' % (algo_name)))
    plt.close()

    return


# generate K arms
n_arms = int(input("How many arms do you want to create?\n"))
n_budgets = int(input("How many budgets do you want to allocate for single simualtion?\n"))
n_reps = int(input("How many replications do you want to conduct?\n"))

##########################
success_rates = [round(random.random(), 3) for i in range(n_arms)]
arms = [BernoulliArm(success_rates[i]) for i in range(n_arms)]

algo_list = {'UCB(0.3)': UCB(0.3), 'BayesBandit': BayesBandit(), 'OCBA': OCBA()}
# algo_list = {'OCBA':OCBA()}
algo = {}
rewards = {}
cpu_time = {}
cumulative_rewards = {}
pcs = {}
for foo in algo_list:
    # algorithm initialization
    algo[foo] = algo_list[foo]
    algo_name = foo
    if foo == 'OCBA':
        init_sim = int((input(
            'How many initial budgets do you want to allocate for each arm in OCBA algorithms with total budget = %d, and %d arms?\n' % (
            n_budgets, n_arms))))
    else:
        init_sim = 0
    # run algorithm
    [rewards[foo], cpu_time[foo], cumulative_rewards[foo], pcs[foo]] = run_algorithm(algo[foo], algo_name, arms,
                                                                                      n_reps, n_budgets, init_sim)
plot(n_arms, n_budgets, n_reps, rewards, cpu_time, cumulative_rewards, pcs)

#########################
n_budgets = int(
    input("Success Rates Variation - How many budgets do you want to allocate for single simualtion?\n"))
base_success_rates = 0.5
increment = 0.1
variation = list(np.array([0.00, 0.25, 0.50, 0.75, 1.00]) * increment + base_success_rates)
algo = {}
for foo in algo_list:
    rewards = {}
    cpu_time = {}
    cumulative_rewards = {}
    pcs = {}
    for var in variation:
        algo[foo] = algo_list[foo]
        algo_name = foo
        success_rates = [base_success_rates for i in range(n_arms)]
        success_rates[0] = var
        arms = [BernoulliArm(success_rates[i]) for i in range(n_arms)]
        [rewards[var], cpu_time[var], cumulative_rewards[var], pcs[var]] = run_algorithm(algo[foo], algo_name, arms,
                                                                                          n_reps, n_budgets, init_sim)
    plot2(foo, variation, n_arms, n_budgets, n_reps, rewards, cpu_time, cumulative_rewards, pcs)

#########################
init_sim = [5, 10, 15, 20, 30]
n_budgets = int(input("OCBA Init-Sim - How many budgets do you want to allocate for single simualtion?\n"))
success_rates = [round(random.random(), 3) for i in range(n_arms)]
arms = [BernoulliArm(success_rates[i]) for i in range(n_arms)]
rewards = {}
cpu_time = {}
cumulative_rewards = {}
pcs = {}
for initsim in init_sim:
    algo_name = 'OCBA'
    [rewards[initsim], cpu_time[initsim], cumulative_rewards[initsim], pcs[initsim]] = run_algorithm(algo_list['OCBA'],
                                                                                                      algo_name, arms,
                                                                                                      n_reps, n_budgets,
                                                                                                      initsim)
plot3(init_sim, n_arms, n_budgets, n_reps, rewards, cpu_time, cumulative_rewards, pcs)


#!/usr/bin/env python
import logging
import os
from qbz95.conf import config_jupyter
from qbz95 import env

deployment = env.params['deployment']

if deployment == 'jupyter':
    config = config_jupyter.ConfigJupyter()
else:
    raise Exception('wrong deployment')

logging.info('deployment = ' + deployment)
logging.info('base_path = ' + config.base_path())
logging.info('data_path = ' + config.data_path())

if config.http_proxy() is not None:
    os.environ['HTTPS_PROXY'] = config.http_proxy()
    os.environ['HTTP_PROXY'] = config.http_proxy()







import gym
env = gym.make('CartPole-v0')

for i_episode in range(20):
    observation = env.reset()
    for t in range(100):
        env.render()
        print(observation)
        action = env.action_space.sample()
        observation, reward, done, info = env.step(action)
        if done:
            print("Episode finished after {} timesteps".format(t+1))
            break



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from collections import OrderedDict
from sklearn.metrics import roc_auc_score, roc_curve, auc, precision_recall_curve, average_precision_score


class ClassificationPredictor(object):
    def __init__(self, model):
        self.model = model

    def predict(self, X):
        pass

    def accuracy(self, X, y):
        pass

    def accuracy_(self, y_pred, y):
        return np.mean(y_pred == y)

    def evaluate(self, X, y):
        pass

    def evaluates(self, X_train, y_train, X_val=None, y_val=None, X_test=None, y_test=None):
        metrics = []
        if X_train is not None:
            metrics_ = self.evaluate(X_train, y_train)
            metrics_ = pd.DataFrame([metrics_.values()], index=["train"], columns=metrics_.keys())
            metrics.append(metrics_)
        if X_val is not None:
            metrics_ = self.evaluate(X_val, y_val)
            metrics_ = pd.DataFrame([metrics_.values()], index=["val"], columns=metrics_.keys())
            metrics.append(metrics_)
        if X_test is not None:
            metrics_ = self.evaluate(X_test, y_test)
            metrics_ = pd.DataFrame([metrics_.values()], index=["test"], columns=metrics_.keys())
            metrics.append(metrics_)
        return pd.concat(metrics)


class MultiClassificationPredictor(ClassificationPredictor):

    def __init__(self, model):
        ClassificationPredictor.__init__(self, model)

    def predict(self, X):
        return np.argmax(self.model.predict(X), axis=-1)

    def accuracy(self, X, y):
        y_pred = self.predict(X)
        y = np.argmax(y, axis=-1)
        return self.accuracy_(y_pred, y)

    def evaluate(self, X, y):
        y_pred = self.predict(X)
        y = np.argmax(y, axis=-1)
        accuracy = self.accuracy_(y_pred, y)
        return {"accuracy":accuracy}


class BinaryClassificationPredictor(ClassificationPredictor):

    def __init__(self, model, cutoff=0.5):
        ClassificationPredictor.__init__(self, model)
        self.cutoff = cutoff

    def predict(self, X):
        return np.int32(self.model.predict(X) > self.cutoff).flatten()

    def accuracy(self, X, y):
        y_pred = self.predict(X)
        return self.accuracy_(y_pred, y)

    def evaluate(self, X, y):
        metrics = OrderedDict()
        y_pred = self.predict(X)
        accuracy = self.accuracy_(y_pred, y)
        metrics["accuracy"] = accuracy

        tp = np.sum((y_pred == 1) * (y == 1)) * 1.0
        fn = np.sum((y_pred == 0) * (y == 1)) * 1.0
        fp = np.sum((y_pred == 1) * (y == 0)) * 1.0
        tn = np.sum((y_pred == 0) * (y == 0)) * 1.0

        recall = tp / (tp + fn) if tp > 0 else 0
        precision = tp / (tp + fp) if tp > 0 else 0
        specificity = tn / (tn + fp) if tn > 0 else 0
        f1 = 2 * recall * precision / (recall + precision) if recall + precision > 0 else 0

        metrics["recall"] = recall
        metrics["precision"] = precision
        metrics["specificity"] = specificity
        metrics["f1"] = f1
        metrics["tp"] = tp
        metrics["fn"] = fn
        metrics["fp"] = fp
        metrics["tn"] = tn

        return metrics


class EvaluateCurve(object):
    def __init__(self, y_true, y_scores):
        y_scores = y_scores.flatten()
        self.fpr, self.tpr, self.roc_thresholds = roc_curve(y_true, y_scores)
        self.auc = auc(self.fpr, self.tpr)

        self.precision, self.recall, self.pr_thresholds = precision_recall_curve(y_true, y_scores)
        self.ap = average_precision_score(y_true, y_scores)

    def plot_roc(self):
        plt.title('ROC Curve [AUC={:.4f}]'.format(self.auc))
        plt.xlim((-0.05, 1.05))
        plt.ylim((-0.05, 1.05))
        plt.ylabel('Sensitivity(TPR)')
        plt.xlabel('1-Specificity(FPR)')
        plt.xticks(np.linspace(0, 1, 11))
        plt.yticks(np.linspace(0, 1, 11))

        plt.plot(self.fpr, self.tpr,  lw=1)

        plt.plot([0,1],[0,1], '--', color='k')
        plt.show()

    def plot_pr(self):
        plt.title('Precision-Recall [AP={:.4f}]'.format(self.ap))
        plt.xlim((0, 1))
        plt.ylim((0, 1.05))
        plt.ylabel('Precision')
        plt.xlabel('Recall')
        plt.xticks(np.linspace(0, 1, 11))
        plt.yticks(np.linspace(0, 1, 11))


        # plt.plot(self.recall, self.precision, lw=1 )
        plt.step(self.recall, self.precision, color='b', alpha=0.2,
                 where='post')
        plt.fill_between(self.recall, self.precision, step='post', alpha=0.2,
                         color='b')
        plt.show()


import numpy as np
import matplotlib.pyplot as plt
from qbz95.rl.bandit.bandit import Bandit, BonuliBandit


class MultiArmedBandit:
    def __init__(self, bandits):
        self.bandits = bandits
        self.returns = []
        self.n = 0
        self.mean_return = 0

    def pull(self):
        j = np.random.choice(len(self.bandits))
        return j, self.bandits[j].pull_update()

    def update(self, ret):
        self.n += 1
        self.returns.append(ret)
        self.mean_return = (1 - 1.0 / self.n) * self.mean_return + 1.0 / self.n * ret[1]

    def pulls(self, n=1):
        for i in range(n):
            self.update(self.pull())
        return self

    def get_cumulative_average(self):
        return np.cumsum([ret for _, ret in self.returns]) / (np.arange(self.n) + 1)

    def plot_return(self, xscale=None, show=True, label='', color="C0"):
        cumulative_average = self.get_cumulative_average()
        plt.plot(cumulative_average, linewidth=1, label=label, color=color)

        for i in range(len(self.bandits)):
            plt.plot(np.ones(self.n) * self.bandits[i].mean, '--', linewidth=1)
        if xscale is not None :
            plt.xscale(xscale)
        if show: plt.show()


def get_bandits(means=[1, 2, 3], mean_return=0):
    return [Bandit(mean=i, mean_return=mean_return) for i in means]


def get_bonuli_bandits(means=[0.2, 0.4, 0.6], mean_return=0):
    return [BonuliBandit(mean=i, mean_return=mean_return) for i in means]


if __name__ == "__main__":
    mab_bandit = MultiArmedBandit(get_bandits()).pulls(10000)
    mab_bandit.plot_return(xscale="log", color="C1")






class Parent:
    parentAttr = 100

    def __init__(self):
        print("parent __init__")

    def parentMethod(self):
        print('parentMethod')

    def setAttr(self, attr):
        Parent.parentAttr = attr

    def getAttr(self):
        print("parent getAttr :", Parent.parentAttr)

    def myMethod(self):
        print('parent myMethod')


class Child(Parent):

    def __init__(self):
        print("child ")

    def childMethod(self):
        print ('childMethod')

    def myMethod(self):
        print ('child myMethod')


def testInherit():
    c = Child()
    c.childMethod()
    c.parentMethod()
    c.setAttr(200)
    c.getAttr()
    c.myMethod()


if __name__ == "__main__":
    testInherit()
    pass



import numpy as np
import logging
import pandas as pd
import matplotlib.pyplot as plt
from qbz95.ml.planar_utils import plot_decision_boundary

#
# function out = mapFeature(X1, X2)
# % MAPFEATURE Feature mapping function to polynomial features
# %
# %   MAPFEATURE(X1, X2) maps the two input features
# %   to quadratic features used in the regularization exercise.
# %
# %   Returns a new feature array with more features, comprising of
# %   X1, X2, X1.^2, X2.^2, X1*X2, X1*X2.^2, etc..
# %
# %   Inputs X1, X2 must be the same size
# %
# degree = 6;
# out = ones(size(X1(:,1)));
# for i = 1:degree
#     for j = 0:i
#         out(:, end+1) = (X1.^(i-j)).*(X2.^j);
#     end
# end
#
# end

def add_intercept(X):
    m = X.shape[1]
    X = np.row_stack((np.ones((1, m)), X))
    return X


def feature_normalize(X):
    mu = np.mean(X, axis=1, keepdims=True)
    sigma = np.std(X, axis=1, keepdims=True, ddof=1)  # Means Delta Degrees of Freedom, need to define a freedom
    X = (X - mu)/sigma
    return X, mu, sigma


def get_logger():
    logger1 = logging.getLogger("ml")
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    logger1.addHandler(handler)
    logger1.setLevel(logging.INFO)
    return logger1


logger = get_logger()


def check_debug_mode(level, epoch_times, print_num_epoch, keep_cost_num_epoch):
    logger.setLevel(level=level)
    if level == logging.DEBUG:
        epoch_times = 1
        print_num_epoch = 1
        keep_cost_num_epoch = 1
    return epoch_times, print_num_epoch, keep_cost_num_epoch


class NumberIterator(object):
    def __init__(self, num_epochs=100):
        assert(num_epochs > 0)
        self.start_flag = False
        self.epoch_times = 0
        self.num_epochs = num_epochs

    def iterate(self):
        if self.start_flag :
            self.epoch_times = self.epoch_times + 1
        else:
            self.start_flag = True
        if self.epoch_times >= self.num_epochs:
            return False
        else:
            return True


class MeanSquaredError(object):

    @classmethod
    def cost(cls, A, Y):
        m = Y.shape[1]
        return np.sum(np.power(A-Y, 2))/2/m

    @classmethod
    def gradient_cost(cls, A, Y):
        m = Y.shape[1]
        return (A-Y)/m


class SigmoidCrossEntropy(object):
    @classmethod
    def cost(cls, A, Y):
        m = Y.shape[1]
        return -np.nansum(np.multiply(Y, np.log(A)) + np.multiply(1-Y, np.log(1-A)))/m

    @classmethod
    def gradient_cost(cls, A, Y):
        m = Y.shape[1]
        return (A-Y)/m


class SoftmaxCrossEntropy(object):
    @classmethod
    def cost(cls, A, Y):
        m = Y.shape[1]
        return -np.sum(np.multiply(Y, np.log(A)))/m

    @classmethod
    def gradient_cost(cls, A, Y):
        m = Y.shape[1]
        return (A-Y)/m


class Linear(object):

    @classmethod
    def activate(cls, Z):
        return Z

    @classmethod
    def gradient(cls, Z):
        A = cls.activate(Z)
        return np.ones(A.shape)


class Relu(object):

    @classmethod
    def activate(cls, Z):
        return np.maximum(0, Z)

    @classmethod
    def gradient(cls, Z):
        A = cls.activate(Z)
        return np.int64(A > 0)


def sigmoid(x, w=None):
    if w is None:
        z = x
    else:
        z = np.dot(x, w)
    return 1.0 / (1.0 + np.exp(-z))


class Sigmoid(object):

    @classmethod
    def activate(cls, Z):
        return sigmoid(Z)

    @classmethod
    def gradient(cls, Z):
        A = cls.activate(Z)
        return np.multiply(A, 1-A)


class Tanh(object):

    @classmethod
    def activate(cls, Z):
        return np.tanh(Z)

    @classmethod
    def gradient(cls, Z):
        A = cls.activate(Z)
        return 1 - np.power(A, 2)

    # @classmethod
    # def gradient_cost(cls, A, Y):
    #     pass


class Softmax(object):
    @classmethod
    def activate(cls, Z):
        E = np.exp(Z)
        return E/np.sum(E, axis = 0, keepdims = True)

    @classmethod
    def gradient(cls, Z):
        A = cls.activate(Z)
        return np.multiply(A, (1-A))

    # @classmethod
    # def compute_cost(cls, A, Y):
    #     m = Y.shape[1]
    #     return -np.sum(np.multiply(Y, np.log(A)))/m

    # @classmethod
    # def gradient_cost(cls, A, Y):
    #     m = Y.shape[1]
    #     return (A-Y)/m


class WeightInitializer(object):

    def get_scale(self, num_neuron, num_weights):
        return 0

    def initial_weights(self, num_neuron, num_weights):
        scale = self.get_scale(num_neuron, num_weights)
        if scale==0:
            return np.zeros((num_neuron, num_weights)), np.zeros((num_neuron, 1))
        else:
            return np.random.randn(num_neuron, num_weights) * scale, np.zeros((num_neuron, 1))


class StaticWeightInitializer(WeightInitializer):

    def __init__(self, scale=0.01):
        self.scale = scale

    def get_scale(self, num_neuron, num_weights):
        return self.scale


class HeWeightInitializer(WeightInitializer):
    def get_scale(self, num_neuron, num_weights):
        return np.sqrt(2.0/num_weights)


class XavierWeightInitializer(WeightInitializer):
    def get_scale(self, num_neuron, num_weights):
        return np.sqrt(1.0/num_weights)


class GbWeightInitializer(WeightInitializer):
    def get_scale(self, num_neuron, num_weights):
        return np.sqrt(2.0/(num_neuron+num_weights))


class Regularizer(object):

    def __init__(self, alpha=0):
        self.alpha = alpha

    def regularize(self, m, W):
        return 0

    def gradient(self, m, W):
        return 0


class L2Regularizer(Regularizer):

    def __init__(self, alpha=0.001):
        Regularizer.__init__(self, alpha)

    def regularize(self, m, W):
        return self.alpha * np.sum(np.multiply(W, W)) / (2 * m)

    def gradient(self, m, W):
        return self.alpha * W / m


class L1Regularizer(Regularizer):

    def __init__(self, alpha=0.001):
        Regularizer.__init__(self, alpha)

    def regularize(self, m, W):
        return self.alpha * np.sum(np.abs(W)) / m

    def gradient(self, m, W):
        return self.alpha * W / np.abs(W) / m


class GradientOptimizer(object):
    def __init__(self, learning_rate=0.01):
        self.learning_rate = learning_rate

    def initialize(self, layer):
        pass

    def update_parameters(self, layer):
        layer.t = layer.t + 1
        layer.W = layer.W - self.learning_rate * layer.dW
        layer.b = layer.b - self.learning_rate * layer.db


class MomentumOptimizer(GradientOptimizer):
    def __init__(self, learning_rate=0.01, beta=0):
        GradientOptimizer.__init__(self, learning_rate)
        assert (beta < 1)
        self.beta = beta

    def initialize(self, layer):
        layer.vW = np.zeros(layer.dW.shape)
        layer.vb = np.zeros(layer.db.shape)

    def update_parameters(self, layer):
        layer.t = layer.t + 1

        layer.vW = (self.beta * layer.vW + (1 - self.beta) * layer.dW)
        layer.vb = (self.beta * layer.vb + (1 - self.beta) * layer.db)
        layer.W = layer.W - self.learning_rate * layer.vW
        layer.b = layer.b - self.learning_rate * layer.vb


# Nesterov Accelerated Gradient
class NagOptimizer(GradientOptimizer):
    pass
    # def __init__(self, learning_rate=0.01, beta=0):
    #     GradientOptimizer.__init__(self, learning_rate)
    #     assert (beta < 1)
    #     self.beta = beta
    #
    # def initialize(self, layer):
    #     layer.vW = np.zeros(layer.dW.shape)
    #     layer.vb = np.zeros(layer.db.shape)
    #
    # def update_parameters(self, layer):
    #     layer.t = layer.t + 1
    #
    #     layer.vW = (self.beta * layer.vW + (1 - self.beta) * layer.dW)
    #     layer.vb = (self.beta * layer.vb + (1 - self.beta) * layer.db)
    #     layer.W = layer.W - self.learning_rate * layer.vW
    #     layer.b = layer.b - self.learning_rate * layer.vb


class AdamOptimizer(object):
    def __init__(self, learning_rate=0.01, beta1=0, beta2=0, epsilon=1e-8):
        GradientOptimizer.__init__(self, learning_rate)
        assert (beta1 < 1)
        assert (beta2 < 1)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon

    def initialize(self, layer):
        layer.vW = np.zeros(layer.dW.shape)
        layer.sW = np.zeros(layer.dW.shape)
        layer.vb = np.zeros(layer.db.shape)
        layer.sb = np.zeros(layer.db.shape)

    def update_parameters(self, layer):
        layer.t = layer.t + 1

        layer.vW = self.beta1 * layer.vW + (1-self.beta1) * layer.dW
        layer.sW = self.beta2 * layer.sW + (1 - self.beta2) * np.power(layer.dW, 2)
        layer.vb = self.beta1 * layer.vb + (1-self.beta1) * layer.db
        layer.sb = self.beta2 * layer.sb + (1 - self.beta2) * np.power(layer.db, 2)

        layer.vW_corrected = layer.vW/(1 - self.beta1 ** layer.t)
        layer.sW_corrected = layer.sW/(1 - self.beta2 ** layer.t)
        layer.vb_corrected = layer.vb/(1 - self.beta1 ** layer.t)
        layer.sb_corrected = layer.sb/(1 - self.beta2 ** layer.t)

        layer.W = layer.W - self.learning_rate*layer.vW_corrected/(np.sqrt(layer.sW_corrected) + self.epsilon)
        layer.b = layer.b - self.learning_rate*layer.vb_corrected/(np.sqrt(layer.sb_corrected) + self.epsilon)





import numpy as np
from qbz95.rl.bandit.multi_armed_bandit import MultiArmedBandit, get_bonuli_bandits
from qbz95.rl.bandit.epsilon_greedy_bandit import EpsilonGreedyBandit
from qbz95.rl.bandit.upper_limit_bandit import UpperLimitBandit
from qbz95.rl.bandit.delay_epsilon_bandit import DelayEpsilonBandit
from qbz95.rl.bandit.ucb_bandit import UCBBandit
import matplotlib.pyplot as plt
import pymc


class ThompsonSamplingBandit(MultiArmedBandit):

    def __init__(self, bandits):
        MultiArmedBandit.__init__(self, bandits)

    def pull(self):
        j = np.argmax([pymc.rbeta(1 + b.win, 1 + b.lost) for b in self.bandits])
        return j, self.bandits[j].pull_update()


if __name__ == "__main__":
    bandit_means = [0.5, 0.6, 0.7, 0.8, 0.9]
    times = 10000
    mab_bandits= [
        UpperLimitBandit(get_bonuli_bandits(bandit_means, mean_return=10)).pulls(times),
        EpsilonGreedyBandit(get_bonuli_bandits(bandit_means), 0.1).pulls(times),
        UCBBandit(get_bonuli_bandits(bandit_means)).pulls(times),
        DelayEpsilonBandit(get_bonuli_bandits(bandit_means)).pulls(times),
        ThompsonSamplingBandit(get_bonuli_bandits(bandit_means)).pulls(times)
    ]

    mab_bandits = sorted(mab_bandits, key=lambda mab_bandit: mab_bandit.mean_return )
    for i in range(len(mab_bandits)):
        mab_bandit = mab_bandits[i]
        label = type(mab_bandit).__name__ + ": {0:.5f}".format(mab_bandit.mean_return)
        mab_bandit.plot_return(xscale="log", show=False, label=label, color='C'+str(i % 8))

    plt.legend()
    plt.show()





import numpy as np


class Initializer:
    def init(self, input_units, output_units):
        return np.random.randn(input_units, output_units)


class ScaleInitializer(Initializer):
    def __init__(self, scale=0.01):
        self.scale = scale

    def init(self, input_units, output_units):
        return np.random.randn(input_units, output_units) * self.scale


class XavierInitializer(Initializer):
    def init(self, input_units, output_units):
        return np.random.randn(input_units, output_units) / np.sqrt(input_units)


class HeInitializer(Initializer):
    def init(self, input_units, output_units):
        return np.random.randn(input_units, output_units) / np.sqrt(input_units / 2)


class GBInitializer(Initializer):
    """Glorot & Bengio"""
    def init(self, input_units, output_units):
        return np.random.randn(input_units, output_units) / np.sqrt((input_units + output_units) / 2)


import json
import os
import numpy as np

from gensim.models import Word2Vec
from gensim.utils import simple_preprocess
from keras.engine import Input
from keras.layers import Embedding, merge
from keras.models import Model

# tokenizer: can change this as needed
tokenize = lambda x: simple_preprocess(x)


def create_embeddings(data_dir,
                      embeddings_path='embeddings.npz',
                      vocab_path='map.json',
                      **params):
    """
    Generate embeddings from a batch of text
    :param embeddings_path: where to save the embeddings
    :param vocab_path: where to save the word-index map
    """

    class SentenceGenerator(object):
        def __init__(self, dirname):
            self.dirname = dirname

        def __iter__(self):
            for fname in os.listdir(self.dirname):
                for line in open(os.path.join(self.dirname, fname)):
                    yield tokenize(line)

    sentences = SentenceGenerator(data_dir)

    model = Word2Vec(sentences, **params)
    weights = model.syn0
    np.save(open(embeddings_path, 'wb'), weights)

    vocab = dict([(k, v.index) for k, v in model.vocab.items()])
    with open(vocab_path, 'w') as f:
        f.write(json.dumps(vocab))


import unittest
from qbz95.ml.utils import *
from qbz95.ml.planar_utils import *
from qbz95.ml.predictor import *
import qbz95.ml.linear_model as lm
from sklearn import metrics

# logger.setLevel(level=logging.DEBUG)


class TestPredictor(unittest.TestCase):
    pd.set_option('display.width', 100)

    def test_ClassificationPredictor_Binary(self):
        np.random.seed(3)
        predictor = ClassificationPredictor()

        X = np.abs(np.random.rand(1, 10))
        Y = np.array([[0, 1, 0, 1, 1, 1, 0, 1, 0, 1]])

        Y_predict = predictor.predict(X)
        print("X={}".format(X))
        print("Y={}".format(Y))
        print("Y_predict={}".format(Y_predict))

        np.testing.assert_array_equal(Y_predict, [[1, 1, 0, 1, 1, 1, 0, 0, 0, 0]])

        accuracy = predictor.accuracy(X, Y)
        print("accuracy={}".format(accuracy))
        np.testing.assert_allclose(accuracy, 0.7)
        metrics = predictor.evaluate(X, Y)
        print("metrics:\n{}".format(metrics))
        np.testing.assert_array_almost_equal(metrics.as_matrix(),
                                             [[0.7,  0, 0.6666667, 0.8, 0.75, 0.7272729,
                                               4,   2,   1,   3]])

    def test_ClassificationPredictor_Multi(self):
        np.random.seed(3)
        predictor = ClassificationPredictor()

        X = np.abs(np.random.rand(2, 10))
        Y = np.array([[0, 1, 0, 1, 1, 1, 0, 1, 0, 1],
                      [1, 0, 1, 0, 0, 0, 1, 0, 1, 0]])

        Y_predict = predictor.predict(X)
        print("X={}".format(X))
        print("Y={}".format(Y))
        print("Y_predict={}".format(Y_predict))

        np.testing.assert_array_equal(Y_predict, [[1, 1, 0, 1, 1, 1, 1, 0, 0, 1],
                                                  [0, 0, 1, 0, 0, 0, 0, 1, 1, 0]])

        accuracy = predictor.accuracy(X, Y)
        print("accuracy={}".format(accuracy))
        np.testing.assert_allclose(accuracy, 0.7)

        metrics = predictor.evaluate(X, Y)
        print("metrics:\n{}".format(metrics))

        np.testing.assert_array_almost_equal(metrics.as_matrix(),
                                             [[0.7, 0, 0.833333,  0.714286,    0.500000, 0.769231,  5,  1,  2,  2],
                                              [0.7, 0, 0.500000,  0.666667,    0.833333, 0.571429,  2,  2,  1,  5]])

    def test_ProbabilityPredictor_Binary(self):
        np.random.seed(3)

        X = np.abs(np.random.rand(1, 10))
        Y = np.array([[0, 1, 0, 1, 1, 1, 0, 1, 0, 1]])
        predictor = ProbabilityPredictor.build(Y)
        np.testing.assert_array_almost_equal(predictor.prior_prob, [[0.6]])

        Y_predict = predictor.predict(X)
        print("X={}".format(X))
        print("Y={}".format(Y))
        print("Y_predict={}".format(Y_predict))

        np.testing.assert_array_equal(Y_predict, [[0, 1, 0, 0, 1, 1, 0, 0, 0, 0]])

        accuracy = predictor.accuracy(X, Y)
        print("accuracy={}".format(accuracy))
        np.testing.assert_allclose(accuracy, 0.7)
        metrics = predictor.evaluate(X, Y)
        print("metrics:\n{}".format(metrics))
        np.testing.assert_array_almost_equal(metrics.as_matrix(),
                                             np.array([[0.7,  0, 0.5, 1.0, 1.0, 0.666667,
                                                        3,   3,   0,   4]]))

    def test_ProbabilityPredictor_Multi(self):
        np.random.seed(3)
        X = np.abs(np.random.rand(3, 10))
        Y = np.array([[0, 1, 0, 1, 1, 1, 0, 1, 0, 0],
                      [1, 0, 1, 0, 0, 0, 0, 0, 1, 0],
                      [0, 0, 0, 0, 0, 0, 1, 0, 0, 1]
                      ])

        predictor = ProbabilityPredictor.build(Y)
        np.testing.assert_array_almost_equal(predictor.prior_prob, [[ 0.5], [ 0.3], [ 0.2]])

        Y_predict = predictor.predict(X)
        print("X={}".format(X))
        print("Y={}".format(Y))
        print("Y_predict={}".format(Y_predict))

        np.testing.assert_array_equal(Y_predict, [[0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                                                  [0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                                                  [1, 1, 1, 0, 1, 1, 1, 0, 1, 1]])

        accuracy = predictor.accuracy(X, Y)
        print("accuracy={}".format(accuracy))
        np.testing.assert_allclose(accuracy, 0.5333333)

        metrics = predictor.evaluate(X, Y)
        print("metrics:\n{}".format(metrics))

        np.testing.assert_array_almost_equal(metrics.as_matrix(),
                                             [[0.533333, 0, 0.2, 1.00, 1.000000, 0.333333,  1,  4,  0,  5],
                                              [0.533333, 0, 0.0, 0.00, 0.857143, 0.000000,  0,  3,  1,  6],
                                              [0.533333, 0, 1.0, 0.25, 0.250000, 0.400000,  2,  0,  6,  2]])


    def test_EvaluateCurve(self):
        train_X_orig, train_Y_orig, test_X_orig, test_Y_orig = load_petal_dataset()
        train_X, train_Y, test_X, test_Y = (train_X_orig, train_Y_orig, test_X_orig, test_Y_orig)
        train_X = map_feature(train_X, 3)
        test_X = map_feature(test_X, 3)
        model = lm.LogisticRegression(num_features=train_X.shape[0], activator=Sigmoid,
                                      regularizer=L2Regularizer(0), iterator=NumberIterator(100),
                                      optimizer=GradientOptimizer(0.5), initializer=WeightInitializer(),
                                      keep_cost_num_epoch=50)

        model.fit(train_X, train_Y, print_cost=False, print_num_epoch=50)
        metrics100 = EvaluateCurve.generate_data(train_X, train_Y, model=model)
        print("metrics100:\n{}".format(metrics100))

        auc = EvaluateCurve.auc(metrics=metrics100)
        sklearn_auc = metrics.auc(1 - metrics100["specificity"], metrics100['recall'])
        print("auc={}".format(auc))
        print("sklearn.auc={}".format(sklearn_auc))
        np.testing.assert_equal(auc, sklearn_auc)

        model.iterator = NumberIterator(200)
        model.fit(train_X, train_Y, print_cost=False, print_num_epoch=50)
        metrics200 = EvaluateCurve.generate_data(train_X, train_Y, model=model)
        print("metrics200:\n{}".format(metrics200))

        model.iterator = NumberIterator(400)
        model.fit(train_X, train_Y, print_cost=False, print_num_epoch=50)
        metrics400 = EvaluateCurve.generate_data(train_X, train_Y, model=model)
        print("metrics400:\n{}".format(metrics400))

        curves = {'NumberIterator(100)': metrics100,
                  'NumberIterator(200)': metrics200,
                  'NumberIterator(400)': metrics400}
        # EvaluateCurve.plot_roc(curves)
        # EvaluateCurve.plot_pr(curves)




